"""Wandb has special data types for logging rich visualizations.

All of the special data types are subclasses of WBValue. All of the data types
    serialize to JSON, since that is what wandb uses to save the objects locally
    and upload them to the W&B server.
"""

from __future__ import print_function

import hashlib
import itertools
import json
import pprint
import shutil
from six.moves import queue
import warnings

import numbers
import collections
import os
import io
import logging
import six
import wandb
import uuid
import json
import codecs
import tempfile
from wandb import util
from wandb.util import has_num
from wandb.compat import tempfile

# Get rid of cleanup warnings in Python 2.7.
warnings.filterwarnings('ignore', 'Implicitly cleaning up', RuntimeWarning, 'wandb.compat.tempfile')

# Staging directory so we can encode raw data into files, then hash them before
# we put them into the Run directory to be uploaded.
MEDIA_TMP = tempfile.TemporaryDirectory('wandb-media')

DATA_FRAMES_SUBDIR = os.path.join('media', 'data_frames')


class WBValue(object):
    """Abstract parent class for things that can be logged by wandb.log() and
        visualized by wandb.

    The objects will be serialized as JSON and always have a _type attribute
        that indicates how to interpret the other fields.

    Returns:
        JSON-friendly `dict` representation of this object that can later be
            serialized to a string.
    """

    def __init__(self):
        pass

    def to_json(self, run):
        """
        """
        raise NotImplementedError


class Histogram(WBValue):
    """
    wandb class for histograms

    This object works just like numpy's histogram function
    https://docs.scipy.org/doc/numpy/reference/generated/numpy.histogram.html

    Examples:
        Generate histogram from a sequence
        ```
        wandb.Histogram([1,2,3])
        ```

        Efficiently initialize from np.histogram.
        ```
        hist = np.histogram(data)
        wandb.Histogram(np_histogram=hist)
        ```

    Arguments:
        sequence (array_like): input data for histogram
        np_histogram (numpy histogram): alternative input of a precoomputed histogram
        num_bins (int): Number of bins for the histogram.  The default number of bins
            is 64.  The maximum number of bins is 512

    Attributes:
        bins ([float]): edges of bins
        histogram ([int]): number of elements falling in each bin
    """
    MAX_LENGTH = 512

    def __init__(self, sequence=None, np_histogram=None, num_bins=64):

        if np_histogram:
            if len(np_histogram) == 2:
                self.histogram = np_histogram[0].tolist() if hasattr(np_histogram[0], 'tolist') else np_histogram[0]
                self.bins = np_histogram[1].tolist() if hasattr(np_histogram[1], 'tolist') else np_histogram[1]
            else:
                raise ValueError(
                    'Expected np_histogram to be a tuple of (values, bin_edges) or sequence to be specified')
        else:
            np = util.get_module(
                "numpy", required="Auto creation of histograms requires numpy")

            self.histogram, self.bins = np.histogram(
                sequence, bins=num_bins)
            self.histogram = self.histogram.tolist()
            self.bins = self.bins.tolist()
        if len(self.histogram) > self.MAX_LENGTH:
            raise ValueError(
                "The maximum length of a histogram is %i" % self.MAX_LENGTH)
        if len(self.histogram) + 1 != len(self.bins):
            raise ValueError("len(bins) must be len(histogram) + 1")

    def to_json(self, run=None):
        return {"_type": "histogram", "values": self.histogram, "bins": self.bins}


class Media(WBValue):
    """A WBValue that we store as a file outside JSON and show in a media panel
    on the front end.

    If necessary, we move or copy the file into the Run's media directory so that it gets
    uploaded.
    """

    def __init__(self, caption=None):
        self._path = None
        # The run under which this object is bound, if any.
        self._run = None
        self._caption = caption

    def _set_file(self, path, is_tmp=False, extension=None):
        self._path = path
        self._is_tmp = is_tmp
        self._extension = extension
        if extension is not None and not path.endswith(extension):
            raise ValueError('Media file extension "{}" must occur at the end of path "{}".'.format(extension, path))

        with open(self._path, 'rb') as f:
            self._sha256 = hashlib.sha256(f.read()).hexdigest()
        self._size = os.path.getsize(self._path)

    @classmethod
    def get_media_subdir(cls):
        raise NotImplementedError

    @classmethod
    def captions(cls, media_items):
        if media_items[0]._caption != None:
            return [m._caption for m in media_items]
        else:
            return False

    def is_bound(self):
        return self._run is not None

    def file_is_set(self):
        return self._path is not None

    def bind_to_run(self, run, key, step, id_=None):
        """Bind this object to a particular Run.

        Calling this function is necessary so that we have somewhere specific to
        put the file associated with this object, from which other Runs can
        refer to it.
        """
        if not self.file_is_set():
            raise AssertionError('bind_to_run called before _set_file')
        if run is None:
            raise TypeError('Argument "run" must not be None.')
        if self.is_bound():
            raise RuntimeError('Value is already bound to a Run: {}'.format(self))
        self._run = run

        # This is a flawed way of checking whether the file is already in
        # the Run directory. It'd be better to check the actual directory
        # components.
        if not os.path.realpath(self._path).startswith(os.path.realpath(self._run.dir)):
            base_path = os.path.join(self._run.dir, self.get_media_subdir())

            if self._extension is None:
                rootname, extension = os.path.splitext(os.path.basename(self._path))
            else:
                extension = self._extension
                rootname = os.path.basename(self._path)[:-len(extension)]

            if self._is_tmp:
                if id_ is None:
                    id_ = self._sha256[:8]

                new_path = os.path.join(base_path, '{}_{}_{}{}'.format(key, step, id_, extension))
                util.mkdir_exists_ok(os.path.dirname(new_path))

                shutil.move(self._path, new_path)

                self._path = new_path
                self._is_tmp = False
            else:
                new_path = os.path.join(base_path, '{}_{}{}'.format(rootname, self._sha256[:8], extension))
                util.mkdir_exists_ok(os.path.dirname(new_path))
                shutil.copy(self._path, new_path)
                self._path = new_path

    def to_json(self, run):
        """Get the JSON-friendly dict that represents this object.

        Only works if `self.bind_to_run()` has previously been called.

        The resulting dict lets you load this object into other W&B runs.
        """
        if not self.is_bound():
            raise RuntimeError(
                'Value of type {} must be bound to a run with bind_to_run() before being serialized to JSON.'.format(type(self).__name__))

        assert self._run is run, "For now we don't support referring to media files across runs."

        return {
            '_type': 'file',  # TODO(adrian): This isn't (yet) a real media type we support on the frontend.
            'path': util.to_forward_slash_path(os.path.relpath(self._path, self._run.dir)),
            'sha256': self._sha256,
            'size': self._size,
            #'entity': self._run.entity,
            #'project': self._run.project_name(),
            #'run': self._run.name,
        }


class BatchableMedia(Media):
    """Parent class for Media we treat specially in batches, like images and
    thumbnails.

    Apart from images, we just use these batches to help organize files by name
    in the media directory.
    """
    @classmethod
    def seq_to_json(self, seq, run, key, step):
        raise NotImplementedError


class Table(Media):
    """This is a table designed to display small sets of records.

    Arguments:
        columns ([str]): Names of the columns in the table.
            Defaults to ["Input", "Output", "Expected"].
        data (array): 2D Array of values that will be displayed as strings.
        dataframe (pandas.DataFrame): DataFrame object used to create the table.
            When set, the other arguments are ignored.
    """
    MAX_ROWS = 10000

    def __init__(self, columns=["Input", "Output", "Expected"], data=None, rows=None, dataframe=None):
        """rows is kept for legacy reasons, we use data to mimic the Pandas api
        """
        super(Table, self).__init__()
        self.columns = columns
        self.data = list(rows or data or [])
        if dataframe is not None:
            assert util.is_pandas_data_frame(dataframe), 'dataframe argument expects a `Dataframe` object'
            self.columns = list(dataframe.columns)
            self.data = []
            for row in range(len(dataframe)):
                self.add_data(*tuple(dataframe[col].values[row] for col in self.columns))

    def add_row(self, *row):
        logging.warning("add_row is deprecated, use add_data")
        self.add_data(*row)

    def add_data(self, *data):
        if len(data) != len(self.columns):
            raise ValueError("This table expects {} columns: {}".format(
                len(self.columns), self.columns))
        self.data.append(list(data))

    def _to_table_json(self):
        # seperate method for testing
        if len(self.data) > Table.MAX_ROWS:
            logging.warn("Truncating wandb.Table object to %i rows." % Table.MAX_ROWS)
        return {"columns": self.columns, "data": self.data[:Table.MAX_ROWS]}

    def bind_to_run(self, *args, **kwargs):
        data = self._to_table_json()
        tmp_path = os.path.join(MEDIA_TMP.name, util.generate_id() + '.table.json')
        data = numpy_arrays_to_lists(data)
        util.json_dump_safer(data, codecs.open(tmp_path, 'w', encoding='utf-8'))
        self._set_file(tmp_path, is_tmp=True, extension='.table.json')
        super(Table, self).bind_to_run(*args, **kwargs)

    @classmethod
    def get_media_subdir(cls):
        return os.path.join('media', 'table')

    def to_json(self, run):
        json_dict = super(Table, self).to_json(run)
        json_dict['_type'] = 'table-file'
        json_dict['ncols'] = len(self.columns)
        json_dict['nrows'] = len(self.data)
        return json_dict


class Audio(BatchableMedia):
    """
        Wandb class for audio clips.

        Args:
            data_or_path (string or numpy array): A path to an audio file
                or a numpy array of audio data.
            sample_rate (int): Sample rate, required when passing in raw
                numpy array of audio data.
            caption (string): Caption to display with audio.
    """

    def __init__(self, data_or_path, sample_rate=None, caption=None):
        """Accepts a path to an audio file or a numpy array of audio data.
        """
        super(Audio, self).__init__()
        self._duration = None
        self._sample_rate = sample_rate
        self._caption = caption

        if isinstance(data_or_path, six.string_types):
            self._set_file(data_or_path, is_tmp=False)
        else:
            if sample_rate == None:
                raise ValueError('Argument "sample_rate" is required when instantiating wandb.Audio with raw data.')

            soundfile = util.get_module(
                "soundfile", required='Raw audio requires the soundfile package. To get it, run "pip install soundfile"')

            tmp_path = os.path.join(MEDIA_TMP.name, util.generate_id() + '.wav')
            soundfile.write(tmp_path, data_or_path, sample_rate)
            self._duration = len(data_or_path) / float(sample_rate)

            self._set_file(tmp_path, is_tmp=True)

    @classmethod
    def get_media_subdir(cls):
        return os.path.join('media', 'audio')

    def to_json(self, run):
        json_dict = super(Audio, self).to_json(run)
        json_dict.update({
            '_type': 'audio-file',
            'sample_rate': self._sample_rate,
            'caption': self._caption,
        })
        return json_dict

    @classmethod
    def seq_to_json(cls, seq, run, key, step):
        audio_list = list(seq)
        for audio in audio_list:
            if not audio.is_bound():
                audio.bind_to_run(run, key, step)

        sf = util.get_module(
            "soundfile", required="wandb.Audio requires the soundfile package. To get it, run: pip install soundfile")
        base_path = os.path.join(run.dir, "media", "audio")
        util.mkdir_exists_ok(base_path)
        meta = {
            "_type": "audio",
            "count": len(audio_list),
            'audio': [a.to_json(run) for a in audio_list],
        }
        sample_rates = cls.sample_rates(audio_list)
        if sample_rates:
            meta["sampleRates"] = sample_rates
        durations = cls.durations(audio_list)
        if durations:
            meta["durations"] = durations
        captions = cls.captions(audio_list)
        if captions:
            meta["captions"] = captions

        return meta

    @classmethod
    def durations(cls, audio_list):
        return [a._duration for a in audio_list]

    @classmethod
    def sample_rates(cls, audio_list):
        return [a._sample_rate for a in audio_list]

    @classmethod
    def captions(cls, audio_list):
        captions = [a._caption for a in audio_list]
        if all(c is None for c in captions):
            return False
        else:
            return ['' if c == None else c for c in captions]


def is_numpy_array(data):
    np = util.get_module(
        "numpy", required="Logging raw point cloud data requires numpy")
    return isinstance(data, np.ndarray)


class Object3D(BatchableMedia):
    """
        Wandb class for 3D point clouds.

        Args:
            data_or_path (numpy array | string | io ):
                Object3D can be initialized from a file or a numpy array.

                The file types supported are obj, gltf, babylon, stl.  You can pass a path to
                    a file or an io object and a file_type which must be one of `'obj', 'gltf', 'babylon', 'stl'`.

                The shape of the numpy array must be one of either:
                ```
                [[x y z],       ...] nx3
                [x y z c],     ...] nx4 where c is a category with supported range [1, 14]
                [x y z r g b], ...] nx4 where is rgb is color
                ```

    """

    SUPPORTED_TYPES = set(['obj', 'gltf', 'babylon', 'stl'])

    def __init__(self, data_or_path, **kwargs):
        super(Object3D, self).__init__()

        if hasattr(data_or_path, 'name'):
            # if the file has a path, we just detect the type and copy it from there
            data_or_path = data_or_path.name

        if hasattr(data_or_path, 'read'):
            if hasattr(data_or_path, 'seek'):
                data_or_path.seek(0)
            object3D = data_or_path.read()

            extension = kwargs.pop("file_type", None)
            if extension == None:
                raise ValueError(
                    "Must pass file type keyword argument when using io objects.")
            if extension not in Object3D.SUPPORTED_TYPES:
                raise ValueError("Object 3D only supports numpy arrays or files of the type: " +
                                 ", ".join(Object3D.SUPPORTED_TYPES))

            tmp_path = os.path.join(MEDIA_TMP.name, util.generate_id() + '.' + extension)
            with open(tmp_path, "w") as f:
                f.write(object3D)

            self._set_file(tmp_path, is_tmp=True)
        elif isinstance(data_or_path, six.string_types):
            path = data_or_path
            try:
                extension = os.path.splitext(data_or_path)[1][1:]
            except:
                raise ValueError(
                    "File type must have an extension")
            if extension not in Object3D.SUPPORTED_TYPES:
                raise ValueError("Object 3D only supports numpy arrays or files of the type: " +
                                 ", ".join(Object3D.SUPPORTED_TYPES))

            self._set_file(data_or_path, is_tmp=False)
        # Supported different types and scene for 3D scenes
        elif 'type' in data_or_path:
            if data_or_path['type'] == 'lidar/beta':
                data = {
                    'type': data_or_path['type'],
                    'vectors': data_or_path['vectors'].tolist(),
                    'points': data_or_path['points'].tolist(),
                    'boxes': data_or_path['boxes'].tolist(),
                }
            else:
                raise ValueError("Type not supported, only 'lidar/beta' is currently supported")

            tmp_path = os.path.join(MEDIA_TMP.name, util.generate_id() + '.pts.json')
            json.dump(data, codecs.open(tmp_path, 'w', encoding='utf-8'),
                      separators=(',', ':'), sort_keys=True, indent=4)
            self._set_file(tmp_path, is_tmp=True, extension='.pts.json')
        elif is_numpy_array(data_or_path):
            data = data_or_path

            if len(data.shape) != 2 or data.shape[1] not in {3, 4, 6}:
                raise ValueError("""The shape of the numpy array must be one of either
                                    [[x y z],       ...] nx3
                                     [x y z c],     ...] nx4 where c is a category with supported range [1, 14]
                                     [x y z r g b], ...] nx4 where is rgb is color""")

            data = data.tolist()
            tmp_path = os.path.join(MEDIA_TMP.name, util.generate_id() + '.pts.json')
            json.dump(data, codecs.open(tmp_path, 'w', encoding='utf-8'),
                      separators=(',', ':'), sort_keys=True, indent=4)
            self._set_file(tmp_path, is_tmp=True, extension='.pts.json')
        else:
            raise ValueError("data must be a numpy or a file object")

    @classmethod
    def get_media_subdir(self):
        return os.path.join('media', 'object3D')

    def to_json(self, run):
        json_dict = super(Object3D, self).to_json(run)
        json_dict['_type'] = 'object3D-file'
        return json_dict

    @classmethod
    def seq_to_json(cls, threeD_list, run, key, step):
        threeD_list = list(threeD_list)
        for i, obj in enumerate(threeD_list):
            if not obj.is_bound():
                obj.bind_to_run(run, key, step, id_=i)

        jsons = [obj.to_json(run) for obj in threeD_list]

        for obj in jsons:
            if not obj['path'].startswith(cls.get_media_subdir()):
                raise ValueError('Files in an array of Object3D\'s must be in the {} directory, not {}'.format(
                    cls.get_media_subdir(), obj['path']))

        return {
            "_type": "object3D",
            "filenames": [os.path.relpath(j['path'], cls.get_media_subdir()) for j in jsons],
            "count": len(jsons),
            'objects': jsons,
        }


class Molecule(BatchableMedia):
    """
        Wandb class for Molecular data

        Args:
            data_or_path ( string | io ):
                Molecule can be initialized from a file name or an io object.
    """

    SUPPORTED_TYPES = set(['pdb', 'pqr', 'mmcif', 'mcif', 'cif', 'sdf', 'sd', 'gro', 'mol2', 'mmtf'])

    def __init__(self, data_or_path, **kwargs):
        super(Molecule, self).__init__(**kwargs)

        if hasattr(data_or_path, 'name'):
            # if the file has a path, we just detect the type and copy it from there
            data_or_path = data_or_path.name

        if hasattr(data_or_path, 'read'):
            if hasattr(data_or_path, 'seek'):
                data_or_path.seek(0)
            molecule = data_or_path.read()

            extension = kwargs.pop("file_type", None)
            if extension == None:
                raise ValueError(
                    "Must pass file type keyword argument when using io objects.")
            if extension not in Molecule.SUPPORTED_TYPES:
                raise ValueError("Molecule 3D only supports files of the type: " +
                                 ", ".join(Molecule.SUPPORTED_TYPES))

            tmp_path = os.path.join(MEDIA_TMP.name, util.generate_id() + '.' + extension)
            with open(tmp_path, "w") as f:
                f.write(molecule)

            self._set_file(tmp_path, is_tmp=True)
        elif isinstance(data_or_path, six.string_types):
            path = data_or_path
            try:
                extension = os.path.splitext(data_or_path)[1][1:]
            except:
                raise ValueError(
                    "File type must have an extension")
            if extension not in Molecule.SUPPORTED_TYPES:
                raise ValueError("Molecule only supports files of the type: " +
                                 ", ".join(Molecule.SUPPORTED_TYPES))

            self._set_file(data_or_path, is_tmp=False)
        else:
            raise ValueError("Data must be file name or a file object")

    @classmethod
    def get_media_subdir(self):
        return os.path.join('media', 'molecule')

    def to_json(self, run):
        json_dict = super(Molecule, self).to_json(run)
        json_dict['_type'] = 'molecule-file'
        if self._caption:
            json_dict['caption'] = self._caption
        return json_dict

    @classmethod
    def seq_to_json(cls, molecule_list, run, key, step):
        molecule_list = list(molecule_list)
        for i, obj in enumerate(molecule_list):
            if not obj.is_bound():
                obj.bind_to_run(run, key, step, id_=i)

        jsons = [obj.to_json(run) for obj in molecule_list]

        for obj in jsons:
            if not obj['path'].startswith(cls.get_media_subdir()):
                raise ValueError('Files in an array of Molecule\'s must be in the {} directory, not {}'.format(
                    cls.get_media_subdir(), obj['path']))

        return {
            "_type": "molecule",
            "filenames": [obj['path'] for obj in jsons],
            "count": len(jsons),
            "captions": Media.captions(molecule_list)
        }


class Html(BatchableMedia):
    """
        Wandb class for arbitrary html

        Arguments:
            data (string or io object): HTML to display in wandb
            inject (boolean): Add a stylesheet to the HTML object.  If set
                to False the HTML will pass through unchanged.
    """

    def __init__(self, data, inject=True):
        super(Html, self).__init__()

        if isinstance(data, str):
            self.html = data
        elif hasattr(data, 'read'):
            if hasattr(data, 'seek'):
                data.seek(0)
            self.html = data.read()
        else:
            raise ValueError("data must be a string or an io object")
        if inject:
            self.inject_head()

        tmp_path = os.path.join(MEDIA_TMP.name, util.generate_id() + '.html')
        with open(tmp_path, 'w') as out:
            print(self.html, file=out)

        self._set_file(tmp_path, is_tmp=True)

    def inject_head(self):
        join = ""
        if "<head>" in self.html:
            parts = self.html.split("<head>", 1)
            parts[0] = parts[0] + "<head>"
        elif "<html>" in self.html:
            parts = self.html.split("<html>", 1)
            parts[0] = parts[0] + "<html><head>"
            parts[1] = "</head>" + parts[1]
        else:
            parts = ["", self.html]
        parts.insert(
            1, '<base target="_blank"><link rel="stylesheet" type="text/css" href="https://app.wandb.ai/normalize.css" />')
        self.html = join.join(parts).strip()

    @classmethod
    def get_media_subdir(self):
        return os.path.join('media', 'html')

    def to_json(self, run):
        json_dict = super(Html, self).to_json(run)
        json_dict['_type'] = 'html-file'
        return json_dict

    @classmethod
    def seq_to_json(cls, html_list, run, key, step):
        base_path = os.path.join(run.dir, cls.get_media_subdir())
        util.mkdir_exists_ok(base_path)
        for i, h in enumerate(html_list):
            if not h.is_bound():
                h.bind_to_run(run, key, step, id_=i)
        meta = {
            "_type": "html",
            "count": len(html_list),
            'html': [h.to_json(run) for h in html_list]
        }
        return meta


class Video(BatchableMedia):

    """
        Wandb representation of video.

        Args:
            data_or_path (numpy array | string | io):
                Video can be initialized with a path to a file or an io object.
                    The format must be "gif", "mp4", "webm" or "ogg".
                    The format must be specified with the format argument.
                Video can be initialized with a numpy tensor.
                    The numpy tensor must be either 4 dimensional or 5 dimensional.
                    Channels should be (time, channel, height, width) or
                        (batch, time, channel, height width)
            caption (string): caption associated with the video for display
            fps (int): frames per second for video. Default is 4.
            format (string): format of video, necessary if initializing with path or io object.
    """

    EXTS = ("gif", "mp4", "webm", "ogg")

    def __init__(self, data_or_path, caption=None, fps=4, format=None):
        super(Video, self).__init__()

        self._fps = fps
        self._format = format or "gif"
        self._width = None
        self._height = None
        self._channels = None
        self._caption = caption
        if self._format not in Video.EXTS:
            raise ValueError("wandb.Video accepts %s formats" % ", ".join(Video.EXTS))

        if isinstance(data_or_path, six.BytesIO):
            filename = os.path.join(MEDIA_TMP.name, util.generate_id() + '.'+ self._format)
            with open(filename, "wb") as f:
                f.write(data_or_path.read())
            self._set_file(filename, is_tmp=True)
        elif isinstance(data_or_path, six.string_types):
            _, ext = os.path.splitext(data_or_path)
            ext = ext[1:].lower()
            if ext not in Video.EXTS:
                raise ValueError("wandb.Video accepts %s formats" % ", ".join(Video.EXTS))
            self._set_file(data_or_path, is_tmp=False)
            #ffprobe -v error -select_streams v:0 -show_entries stream=width,height -of csv=p=0 data_or_path
        else:
            if hasattr(data_or_path, "numpy"):  # TF data eager tensors
                self.data = data_or_path.numpy()
            elif is_numpy_array(data_or_path):
                self.data = data_or_path
            else:
                raise ValueError("wandb.Video accepts a file path or numpy like data as input")
            self.encode()

    def encode(self):
        mpy = util.get_module(
            "moviepy.editor", required='wandb.Video requires moviepy and imageio when passing raw data.  Install with "pip install moviepy imageio"')
        tensor = self._prepare_video(self.data)
        _, self._height, self._width, self._channels = tensor.shape

        # encode sequence of images into gif string
        clip = mpy.ImageSequenceClip(list(tensor), fps=self._fps)

        filename = os.path.join(MEDIA_TMP.name, util.generate_id() + '.'+ self._format)
        try:  # older version of moviepy does not support progress_bar argument.
            if self._format == "gif":
                clip.write_gif(filename, verbose=False, progress_bar=False)
            else:
                clip.write_videofile(filename, verbose=False, progress_bar=False)
        except TypeError:
            if self._format == "gif":
                clip.write_gif(filename, verbose=False)
            else:
                clip.write_videofile(filename, verbose=False)
        self._set_file(filename, is_tmp=True)

    @classmethod
    def get_media_subdir(cls):
        return os.path.join('media', 'videos')

    def to_json(self, run):
        json_dict = super(Video, self).to_json(run)
        json_dict['_type'] = 'video-file'

        if self._width is not None:
            json_dict['width'] = self._width
        if self._height is not None:
            json_dict['height'] = self._height
        if self._caption:
            json_dict['caption'] = self._caption

        return json_dict

    def _prepare_video(self, V):
        """This logic was mostly taken from tensorboardX"""
        np = util.get_module(
            "numpy", required='wandb.Video requires numpy when passing raw data. To get it, run "pip install numpy".')
        if V.ndim < 4:
            raise ValueError("Video must be atleast 4 dimensions: time, channels, height, width")
        if V.ndim == 4:
            V = V.reshape(1, *V.shape)
        b, t, c, h, w = V.shape

        if V.dtype != np.uint8:
            logging.warning("Converting video data to uint8")
            V = V.astype(np.uint8)

        def is_power2(num):
            return num != 0 and ((num & (num - 1)) == 0)

        # pad to nearest power of 2, all at once
        if not is_power2(V.shape[0]):
            len_addition = int(2**V.shape[0].bit_length() - V.shape[0])
            V = np.concatenate(
                (V, np.zeros(shape=(len_addition, t, c, h, w))), axis=0)

        n_rows = 2**((b.bit_length() - 1) // 2)
        n_cols = V.shape[0] // n_rows

        V = np.reshape(V, newshape=(n_rows, n_cols, t, c, h, w))
        V = np.transpose(V, axes=(2, 0, 4, 1, 5, 3))
        V = np.reshape(V, newshape=(t, n_rows * h, n_cols * w, c))
        return V

    @classmethod
    def seq_to_json(cls, videos, run, key, step):
        base_path = os.path.join(run.dir, cls.get_media_subdir())
        util.mkdir_exists_ok(base_path)
        for i, v in enumerate(videos):
            if not v.is_bound():
                v.bind_to_run(run, key, step, id_=i)
        meta = {
            "_type": "videos",
            "count": len(videos),
            'videos': [v.to_json(run) for v in videos],
            "captions": Video.captions(videos)
        }
        return meta

    @classmethod
    def captions(cls, videos):
        if videos[0]._caption != None:
            return [v._caption for v in videos]
        else:
            return False


class Image(BatchableMedia):
    """
        Wandb class for images.

        Args:
            data_or_path (numpy array | string | io): Accepts numpy array of
                image data, or a PIL image. The class attempts to infer
                the data format and converts it.
            mode (string): The PIL mode for an image. Most common are "L", "RGB",
                "RGBA". Full explanation at https://pillow.readthedocs.io/en/4.2.x/handbook/concepts.html#concept-modes.
            caption (string): Label for display of image.
    """

    MAX_THUMBNAILS = 108

    # PIL limit
    MAX_DIMENSION = 65500

    def __init__(self, data_or_path, mode=None, caption=None, grouping=None, boxes=None, masks=None):
        super(Image, self).__init__()
        # TODO: We should remove grouping, it's a terrible name and I don't
        # think anyone uses it.

        self._grouping = grouping
        self._caption = caption
        self._width = None
        self._height = None
        self._image = None
        
        self._boxes = None 
        if boxes: 
            if not isinstance(boxes, dict):
                raise ValueError("Images \"boxes\" argument must be a dictionary")
            boxes_final = {}
            for key in boxes:
                boxes_final[key] = BoundingBoxes2D(boxes[key], key)
            self._boxes = boxes_final

        self._masks = None
        if masks:
            if not isinstance(masks, dict):
                raise ValueError("Images \"masks\" argument must be a dictionary")
            masks_final = {}
            for key in masks:
                masks_final[key] = ImageMask(masks[key], key)
            self._masks = masks_final

        PILImage = util.get_module(
            "PIL.Image", required='wandb.Image needs the PIL package. To get it, run "pip install pillow".')

        if isinstance(data_or_path, six.string_types):
            self._set_file(data_or_path, is_tmp=False)
            self._image = PILImage.open(data_or_path)
        else:
            data = data_or_path

            if util.is_matplotlib_typename(util.get_full_typename(data)):
                buf = six.BytesIO()
                util.ensure_matplotlib_figure(data).savefig(buf)
                self._image = PILImage.open(buf)
            elif isinstance(data, PILImage.Image):
                self._image = data
            elif util.is_pytorch_tensor_typename(util.get_full_typename(data)):
                vis_util = util.get_module(
                    "torchvision.utils", "torchvision is required to render images")
                if hasattr(data, "requires_grad") and data.requires_grad:
                    data = data.detach()
                data = vis_util.make_grid(data, normalize=True)
                self._image = PILImage.fromarray(data.mul(255).clamp(
                    0, 255).byte().permute(1, 2, 0).cpu().numpy())
            else:
                if hasattr(data, "numpy"):  # TF data eager tensors
                    data = data.numpy()
                if data.ndim > 2:
                    data = data.squeeze()  # get rid of trivial dimensions as a convenience
                self._image = PILImage.fromarray(
                    self.to_uint8(data), mode=mode or self.guess_mode(data))


            tmp_path = os.path.join(MEDIA_TMP.name, util.generate_id() + '.png')
            self._image.save(tmp_path, transparency=None)
            self._set_file(tmp_path, is_tmp=True)

        self._width, self._height = self._image.size

    @classmethod
    def get_media_subdir(cls):
        return os.path.join('media', 'images')

    def bind_to_run(self, *args, **kwargs):
        super(Image, self).bind_to_run(*args, **kwargs)
        if self._boxes is not None:
            for k in self._boxes:
                self._boxes[k].bind_to_run(*args, **kwargs)

        if self._masks is not None:
            for k in self._masks:
                self._masks[k].bind_to_run(*args, **kwargs)

    def to_json(self, run):
        json_dict = super(Image, self).to_json(run)
        json_dict['_type'] = 'image-file'
        json_dict['format'] = "png"

        if self._width is not None:
            json_dict['width'] = self._width
        if self._height is not None:
            json_dict['height'] = self._height
        if self._grouping:
            json_dict['grouping'] = self._grouping
        if self._caption:
            json_dict['caption'] = self._caption
        if self._boxes:
            json_dict['boxes'] = {
                    k: box.to_json(run) for (k, box) in self._boxes.items()}
        if self._masks:
            json_dict['masks'] = {
                k: mask.to_json(run) for (k, mask) in self._masks.items()}

        return json_dict

    def guess_mode(self, data):
        """
        Guess what type of image the np.array is representing
        """
        # TODO: do we want to support dimensions being at the beginning of the array?
        if data.ndim == 2:
            return "L"
        elif data.shape[-1] == 3:
            return "RGB"
        elif data.shape[-1] == 4:
            return "RGBA"
        else:
            raise ValueError(
                "Un-supported shape for image conversion %s" % list(data.shape))

    @classmethod
    def to_uint8(self, data):
        """
        Converts floating point image on the range [0,1] and integer images
        on the range [0,255] to uint8, clipping if necessary.
        """
        np = util.get_module(
            "numpy", required="wandb.Image requires numpy if not supplying PIL Images: pip install numpy")

        # I think it's better to check the image range vs the data type, since many
        # image libraries will return floats between 0 and 255

        # some images have range -1...1 or 0-1
        dmin = np.min(data)
        if dmin < 0:
            data = (data - np.min(data)) / np.ptp(data)
        if np.max(data) <= 1.0:
            data = (data * 255).astype(np.int32)

        #assert issubclass(data.dtype.type, np.integer), 'Illegal image format.'
        return data.clip(0, 255).astype(np.uint8)

    @classmethod
    def seq_to_json(cls, images, run, key, step):
        """
        Combines a list of images into a single sprite returning meta information
        """
        from PIL import Image as PILImage
        base = os.path.join(run.dir, cls.get_media_subdir())
        width, height = images[0]._image.size
        num_images_to_log = len(images)

        if num_images_to_log > Image.MAX_THUMBNAILS:
            logging.warning(
                "Only %i images will be uploaded." % Image.MAX_THUMBNAILS)
            num_images_to_log = Image.MAX_THUMBNAILS
            images = images[:num_images_to_log]

        if width * num_images_to_log > Image.MAX_DIMENSION:
            max_images_by_dimension = Image.MAX_DIMENSION // width
            logging.warning('Only {} images will be uploaded. The maximum total width for a set of thumbnails is 65,500px, or {} images, each with a width of {} pixels.'.format(
                max_images_by_dimension, max_images_by_dimension, width))
            num_images_to_log = max_images_by_dimension

        total_width = width * num_images_to_log
        sprite = PILImage.new(
            mode='RGB',
            size=(total_width, height),
            color=(0, 0, 0))
        for i, image in enumerate(images[:num_images_to_log]):
            location = width * i
            sprite.paste(image._image, (location, 0))
        fname = '{}_{}.png'.format(key, step)
        # fname may contain a slash so we create the directory
        util.mkdir_exists_ok(os.path.dirname(os.path.join(base, fname)))
        sprite.save(os.path.join(base, fname), transparency=None)
        meta = {"width": width, "height": height, "format": "png",
                "count": num_images_to_log, "_type": "images"}
        # TODO: hacky way to enable image grouping for now
        grouping = images[0]._grouping
        if grouping:
            meta["grouping"] = grouping

        captions = Image.all_captions(images)

        if captions:
            meta["captions"] = captions

        all_masks = Image.all_masks(images, run, key, step)

        if all_masks:
            meta["all_masks"] = all_masks

        all_boxes = Image.all_boxes(images, run, key, step)

        if all_boxes:
            meta["all_boxes"] = all_boxes

        return meta

    @classmethod
    def all_masks(cls, images, run, run_key, step):
        all_mask_groups = []
        for image in images:
            if image._masks:
                mask_group = {}
                for k in image._masks:
                    mask = image._masks[k]
                    mask.bind_to_run(run, run_key, step)
                    mask_group[k] = mask.to_json(run)
                all_mask_groups.append(mask_group)
            else:
                all_mask_groups.append(None)
        if all_mask_groups and not all(x is None for x in all_mask_groups):
            return all_mask_groups
        else:
            return False

    @classmethod
    def all_boxes(cls, images, run, run_key, step):
        all_box_groups = []
        for image in images:
            if image._boxes:
                box_group = {}
                for k in image._boxes:
                    box = image._boxes[k]
                    box.bind_to_run(run, run_key, step)
                    box_group[k] = box.to_json(run)
                all_box_groups.append(box_group)
            else:
                all_box_groups.append(None)
        if all_box_groups and not all(x is None for x in all_box_groups):
            return all_box_groups
        else: 
            return False

    @classmethod
    def all_captions(cls, images ):
        if images[0]._caption != None:
            return [i._caption for i in images]
        else:
            return False

# Allows encoding of arbitrary JSON structures
# as a file
#
# This class should be used as an abstract class
# extended to have validation methods


class JSONMetadata(Media):
    """
    JSONMetadata is a type for encoding arbitrary metadata as files.
    """

    def __init__(self, val, **kwargs):
        super(JSONMetadata, self).__init__()

        self.validate(val)
        self._val = val

        ext = "." + self.type_name() + ".json"
        tmp_path = os.path.join(MEDIA_TMP.name, util.generate_id() + ext)
        util.json_dump_uncompressed(self._val, codecs.open(tmp_path, 'w', encoding='utf-8'))
        self._set_file(tmp_path, is_tmp=True, extension=ext)

    def get_media_subdir(self):
        return os.path.join('media', 'metadata', self.type_name())

    def to_json(self, run):
        json_dict = super(JSONMetadata, self).to_json(run)
        json_dict['_type'] = self.type_name()

        return json_dict

    # These methods should be overridden in the child class
    def type_name(self):
        return "metadata"

    def validate(self, val):
        return True


class BoundingBoxes2D(JSONMetadata):
    """
    Wandb class for 2D bounding Boxes
    """

    def __init__(self, val, key, **kwargs):
        super(BoundingBoxes2D, self).__init__(val)
        self._val = val["box_data"]
        self._key = key
        # Add default class mapping
        if not "class_labels" in val:
            np = util.get_module("numpy", required="Semantic Segmentation mask support requires numpy")
            classes = np.unique(list(map( lambda box: box["class_id"], val["box_data"]))).astype(np.int32).tolist()
            class_labels = dict((c, "class_" + str(c)) for c in classes)
            self._class_labels = class_labels
        else:
            self._class_labels = val["class_labels"]

    def bind_to_run(self, run, key, step):
        # bind_to_run key argument is the Image parent key
        # the self._key value is the mask's sub key
        super(BoundingBoxes2D, self).bind_to_run(run, key, step)
        run._add_singleton("bounding_box/class_labels", key + "_wandb_delimeter_" + self._key , self._class_labels)


    def type_name(self):
        return "boxes2D"

    def validate(self, val):
        # Optional argument
        if ("class_labels" in val):
            for k, v in list(val["class_labels"].items()):
                if (not isinstance(k, numbers.Number)) or (not isinstance(v, six.string_types)):
                    raise TypeError("Class labels must be a dictionary of numbers to string")

        boxes = val['box_data']
        if not isinstance(boxes, collections.Sequence):
            raise TypeError("Boxes must be a list")

        for box in boxes:
            # Required arguments
            error_str = "Each box must contain a position with: middle, width, and height or \
                    \nminX, maxX, minY, maxY."
            if not "position" in box:
                raise TypeError(error_str)
            else:
                valid = False
                if "middle" in box["position"] and len(box["position"]["middle"]) == 2 and \
                   has_num(box["position"], "width") and \
                   has_num(box["position"], "height"):
                    valid = True
                elif has_num(box["position"], "minX") and \
                        has_num(box["position"], "maxX") and \
                        has_num(box["position"], "minY") and \
                        has_num(box["position"], "maxY"):
                    valid = True

                if not valid:
                    raise TypeError(error_str)

            # Optional arguments
            if ("scores" in box) and not isinstance(box["scores"], dict):
                raise TypeError("Box scores must be a dictionary")
            elif ("scores" in box):
                for k, v in list(box["scores"].items()):
                    if not isinstance(k, six.string_types):
                        raise TypeError("A score key must be a string")
                    if not isinstance(v, numbers.Number):
                        raise TypeError("A score value must be a number")

            if ("class_id" in box) and not isinstance(box["class_id"], six.integer_types):
                raise TypeError("A box's class_id must be an integer")

            # Optional
            if ("box_caption" in box) and not isinstance(box["box_caption"], six.string_types):
                raise TypeError("A box's caption must be a string")


class ImageMask(Media):
    """
    Wandb class for image masks, useful for segmentation tasks
    """

    def __init__(self, val, key, **kwargs):
        super(ImageMask, self).__init__()

        np = util.get_module("numpy", required="Semantic Segmentation mask support requires numpy")
        # Add default class mapping
        if not "class_labels" in val:
            classes = np.unique(val["mask_data"]).astype(np.int32).tolist()
            class_labels = dict((c, "class_" + str(c)) for c in classes)
            val["class_labels"] = class_labels

        self.validate(val)
        self._val = val
        self._key = key

        ext = "." + self.type_name() + ".png"
        tmp_path = os.path.join(MEDIA_TMP.name, util.generate_id() + ext)

        PILImage = util.get_module(
            "PIL.Image", required='wandb.Image needs the PIL package. To get it, run "pip install pillow".')
        image = PILImage.fromarray(val["mask_data"].astype(np.int8), mode="L")

        image.save(tmp_path, transparency=None)
        self._set_file(tmp_path, is_tmp=True, extension=ext)

    def bind_to_run(self, run, key, step):
        # bind_to_run key argument is the Image parent key
        # the self._key value is the mask's sub key
        super(ImageMask, self).bind_to_run(run, key, step)
        class_labels = self._val["class_labels"]

        run._add_singleton("mask/class_labels", key + "_wandb_delimeter_" + self._key , class_labels)

    def get_media_subdir(self):
        return os.path.join('media', 'images', self.type_name())

    def to_json(self, run):
        json_dict = super(ImageMask, self).to_json(run)
        json_dict['_type'] = self.type_name()

        return json_dict

    def type_name(self):
        return "mask"

    def validate(self, mask):
        np = util.get_module("numpy", required="Semantic Segmentation mask support requires numpy")

        # 2D Make this work with all tensor(like) types
        if not "mask_data" in mask:
            raise TypeError("Missing key \"mask_data\": A mask requires mask data(A 2D array representing the predctions)")
        else:
            error_str = "mask_data must be a 2d array"
            shape = mask["mask_data"].shape
            if len(shape) != 2:
                raise TypeError(error_str)
            if not ((mask["mask_data"] >= 0).all() and (mask["mask_data"] <= 255).all()) and issubclass(mask.dtype.type, np.integer):
                raise TypeError("Mask data must be integers between 0 and 255")

        # Optional argument
        if ("class_labels" in mask):
            for k, v in list(mask["class_labels"].items()):
                if (not isinstance(k, numbers.Number)) or (not isinstance(v, six.string_types)):
                    raise TypeError("Class labels must be a dictionary of numbers to string")


class Plotly(Media):
    """
        Wandb class for plotly plots.

        Args:
            val: matplotlib or plotly figure
    """

    @classmethod
    def make_plot_media(cls, val):
        if util.is_matplotlib_typename(util.get_full_typename(val)):
            val = util.ensure_matplotlib_figure(val)
            if any(len(ax.images) > 0 for ax in val.axes):
                return Image(val)
            val = util.ensure_matplotlib_figure(val)
            # otherwise, convert to plotly figure
            tools = util.get_module(
                "plotly.tools", required="plotly is required to log interactive plots, install with: pip install plotly or convert the plot to an image with `wandb.Image(plt)`")
            val = tools.mpl_to_plotly(val)
        return cls(val)

    def __init__(self, val, **kwargs):
        super(Plotly, self).__init__()
        if not util.is_plotly_figure_typename(util.get_full_typename(val)):
            raise ValueError(
                'Logged plots must be plotly figures, or matplotlib plots convertible to plotly via mpl_to_plotly')

        tmp_path = os.path.join(MEDIA_TMP.name, util.generate_id() + '.plotly.json')
        val = numpy_arrays_to_lists(val.to_plotly_json())
        util.json_dump_safer(val, codecs.open(tmp_path, 'w', encoding='utf-8'))
        self._set_file(tmp_path, is_tmp=True, extension='.plotly.json')

    def get_media_subdir(self):
        return os.path.join('media', 'plotly')

    def to_json(self, run):
        json_dict = super(Plotly, self).to_json(run)
        json_dict['_type'] = 'plotly-file'
        return json_dict


class Graph(Media):
    """Wandb class for graphs

    This class is typically used for saving and diplaying neural net models.  It
    represents the graph as an array of nodes and edges.  The nodes can have
    labels that can be visualized by wandb.

    Examples:
        Import a keras model:
        ```
            Graph.from_keras(keras_model)
        ```

    Attributes:
        format (string): Format to help wandb display the graph nicely.
        nodes ([wandb.Node]): List of wandb.Nodes
        nodes_by_id (dict): dict of ids -> nodes
        edges ([(wandb.Node, wandb.Node)]): List of pairs of nodes interpreted as edges
        loaded (boolean): Flag to tell whether the graph is completely loaded
        root (wandb.Node): root node of the graph
    """

    def __init__(self, format="keras"):
        super(Graph, self).__init__()
        # LB: TODO: I think we should factor criterion and criterion_passed out
        self.format = format
        self.nodes = []
        self.nodes_by_id = {}
        self.edges = []
        self.loaded = False
        self.criterion = None
        self.criterion_passed = False
        self.root = None  # optional root Node if applicable

    def _to_graph_json(self, run=None):
        # Needs to be it's own function for tests
        return {"format": self.format,
                "nodes": [node.to_json() for node in self.nodes],
                "edges": [edge.to_json() for edge in self.edges]}

    def bind_to_run(self, *args, **kwargs):
        data = self._to_graph_json()
        tmp_path = os.path.join(MEDIA_TMP.name, util.generate_id() + '.graph.json')
        data = numpy_arrays_to_lists(data)
        util.json_dump_safer(data, codecs.open(tmp_path, 'w', encoding='utf-8'))
        self._set_file(tmp_path, is_tmp=True, extension='.graph.json')
        if self.is_bound():
            return
        super(Graph, self).bind_to_run(*args, **kwargs)

    @classmethod
    def get_media_subdir(cls):
        return os.path.join('media', 'graph')

    def to_json(self, run):
        json_dict = super(Graph, self).to_json(run)
        json_dict['_type'] = 'graph-file'
        return json_dict

    def __getitem__(self, nid):
        return self.nodes_by_id[nid]

    def pprint(self):
        for edge in self.edges:
            pprint.pprint(edge.attributes)
        for node in self.nodes:
            pprint.pprint(node.attributes)

    def add_node(self, node=None, **node_kwargs):
        if node is None:
            node = Node(**node_kwargs)
        elif node_kwargs:
            raise ValueError('Only pass one of either node ({node}) or other keyword arguments ({node_kwargs})'.format(
                node=node, node_kwargs=node_kwargs))
        self.nodes.append(node)
        self.nodes_by_id[node.id] = node

        return node

    def add_edge(self, from_node, to_node):
        edge = Edge(from_node, to_node)
        self.edges.append(edge)

        return edge

    @classmethod
    def from_keras(cls, model):
        graph = cls()
        # Shamelessly copied from keras/keras/utils/layer_utils.py

        if model.__class__.__name__ == 'Sequential':
            sequential_like = True
        elif not hasattr(model, "_is_graph_network") or not model._is_graph_network:
            # We treat subclassed models as a simple sequence of layers,
            # for logging purposes.
            sequential_like = True
        else:
            sequential_like = True
            nodes_by_depth = model._nodes_by_depth.values()
            nodes = []
            for v in nodes_by_depth:
                # TensorFlow2 doesn't insure inbound is always a list
                inbound = v[0].inbound_layers
                if not hasattr(inbound, '__len__'):
                    inbound = [inbound]
                if (len(v) > 1) or (len(v) == 1 and len(inbound) > 1):
                    # if the model has multiple nodes
                    # or if the nodes have multiple inbound_layers
                    # the model is no longer sequential
                    sequential_like = False
                    break
                nodes += v
            if sequential_like:
                # search for shared layers
                for layer in model.layers:
                    flag = False
                    if hasattr(layer, "_inbound_nodes"):
                        for node in layer._inbound_nodes:
                            if node in nodes:
                                if flag:
                                    sequential_like = False
                                    break
                                else:
                                    flag = True
                    if not sequential_like:
                        break

        relevant_nodes = None
        if sequential_like:
            # header names for the different log elements
            to_display = ['Layer (type)', 'Output Shape', 'Param #']
        else:
            relevant_nodes = []
            for v in model._nodes_by_depth.values():
                relevant_nodes += v

        layers = model.layers
        for i in range(len(layers)):
            node = Node.from_keras(layers[i])
            if hasattr(layers[i], '_inbound_nodes'):
                for in_node in layers[i]._inbound_nodes:
                    if relevant_nodes and in_node not in relevant_nodes:
                        # node is not part of the current network
                        continue
                    for in_layer in nest(in_node.inbound_layers):
                        inbound_keras_node = Node.from_keras(in_layer)

                        if (inbound_keras_node.id not in graph.nodes_by_id):
                            graph.add_node(inbound_keras_node)
                        inbound_node = graph.nodes_by_id[inbound_keras_node.id]

                        graph.add_edge(inbound_node, node)
            graph.add_node(node)
        return graph


class Node(WBValue):
    """
    Node used in :obj:`Graph`
    """

    def __init__(self, id=None, name=None, class_name=None, size=None, parameters=None, output_shape=None, is_output=None, num_parameters=None, node=None):
        self._attributes = {'name': None}
        self.in_edges = {}  # indexed by source node id
        self.out_edges = {}  # indexed by dest node id
        # optional object (eg. PyTorch Parameter or Module) that this Node represents
        self.obj = None

        if node is not None:
            self._attributes.update(node._attributes)
            del self._attributes['id']
            self.obj = node.obj

        if id is not None:
            self.id = id
        if name is not None:
            self.name = name
        if class_name is not None:
            self.class_name = class_name
        if size is not None:
            self.size = size
        if parameters is not None:
            self.parameters = parameters
        if output_shape is not None:
            self.output_shape = output_shape
        if is_output is not None:
            self.is_output = is_output
        if num_parameters is not None:
            self.num_parameters = num_parameters

    def to_json(self, run=None):
        return self._attributes

    def __repr__(self):
        return repr(self._attributes)

    @property
    def id(self):
        """Must be unique in the graph"""
        return self._attributes.get('id')

    @id.setter
    def id(self, val):
        self._attributes['id'] = val
        return val

    @property
    def name(self):
        """Usually the type of layer or sublayer"""
        return self._attributes.get('name')

    @name.setter
    def name(self, val):
        self._attributes['name'] = val
        return val

    @property
    def class_name(self):
        """Usually the type of layer or sublayer"""
        return self._attributes.get('class_name')

    @class_name.setter
    def class_name(self, val):
        self._attributes['class_name'] = val
        return val

    @property
    def functions(self):
        return self._attributes.get('functions', [])

    @functions.setter
    def functions(self, val):
        self._attributes["functions"] = val
        return val

    @property
    def parameters(self):
        return self._attributes.get('parameters', [])

    @parameters.setter
    def parameters(self, val):
        self._attributes["parameters"] = val
        return val

    @property
    def size(self):
        return self._attributes.get('size')

    @size.setter
    def size(self, val):
        """Tensor size"""
        self._attributes['size'] = tuple(val)
        return val

    @property
    def output_shape(self):
        return self._attributes.get('output_shape')

    @output_shape.setter
    def output_shape(self, val):
        """Tensor output_shape"""
        self._attributes['output_shape'] = val
        return val

    @property
    def is_output(self):
        return self._attributes.get('is_output')

    @is_output.setter
    def is_output(self, val):
        """Tensor is_output"""
        self._attributes['is_output'] = val
        return val

    @property
    def num_parameters(self):
        return self._attributes.get('num_parameters')

    @num_parameters.setter
    def num_parameters(self, val):
        """Tensor num_parameters"""
        self._attributes['num_parameters'] = val
        return val

    @property
    def child_parameters(self):
        return self._attributes.get('child_parameters')

    @child_parameters.setter
    def child_parameters(self, val):
        """Tensor child_parameters"""
        self._attributes['child_parameters'] = val
        return val

    @property
    def is_constant(self):
        return self._attributes.get('is_constant')

    @is_constant.setter
    def is_constant(self, val):
        """Tensor is_constant"""
        self._attributes['is_constant'] = val
        return val

    @classmethod
    def from_keras(cls, layer):
        node = cls()

        try:
            output_shape = layer.output_shape
        except AttributeError:
            output_shape = ['multiple']

        node.id = layer.name
        node.name = layer.name
        node.class_name = layer.__class__.__name__
        node.output_shape = output_shape
        node.num_parameters = layer.count_params()

        return node


class Edge(WBValue):
    """
    Edge used in :obj:`Graph`
    """

    def __init__(self, from_node, to_node):
        self._attributes = {}
        self.from_node = from_node
        self.to_node = to_node

    def __repr__(self):
        temp_attr = dict(self._attributes)
        del temp_attr['from_node']
        del temp_attr['to_node']
        temp_attr['from_id'] = self.from_node.id
        temp_attr['to_id'] = self.to_node.id
        return str(temp_attr)

    def to_json(self, run=None):
        return [self.from_node.id, self.to_node.id]

    @property
    def name(self):
        """Optional, not necessarily unique"""
        return self._attributes.get('name')

    @name.setter
    def name(self, val):
        self._attributes['name'] = val
        return val

    @property
    def from_node(self):
        return self._attributes.get('from_node')

    @from_node.setter
    def from_node(self, val):
        self._attributes['from_node'] = val
        return val

    @property
    def to_node(self):
        return self._attributes.get('to_node')

    @to_node.setter
    def to_node(self, val):
        self._attributes['to_node'] = val
        return val


def nest(thing):
    # Use tensorflows nest function if available, otherwise just wrap object in an array"""

    tfutil = util.get_module('tensorflow.python.util')
    if tfutil:
        return tfutil.nest.flatten(thing)
    else:
        return [thing]


def history_dict_to_json(run, payload, step=None):
    # Converts a History row dict's elements so they're friendly for JSON serialization.

    if step is None:
        # We should be at the top level of the History row; assume this key is set.
        step = payload['_step']

    # We use list here because we were still seeing cases of RuntimeError dict changed size
    for key in list(payload):
        val = payload[key]
        if isinstance(val, dict):
            payload[key] = history_dict_to_json(run, val, step=step)
        else:
            payload[key] = val_to_json(run, key, val, step=step)

    return payload


def numpy_arrays_to_lists(payload):
    # Casts all numpy arrays to lists so we don't convert them to histograms, primarily for Plotly

    if isinstance(payload, dict):
        res = {}
        for key, val in six.iteritems(payload):
            res[key] = numpy_arrays_to_lists(val)
        return res
    elif isinstance(payload, collections.Sequence) and not isinstance(payload, six.string_types):
        return [numpy_arrays_to_lists(v) for v in payload]
    elif util.is_numpy_array(payload):
        return [numpy_arrays_to_lists(v) for v in payload.tolist()]

    return payload


def val_to_json(run, key, val, step='summary'):
    # Converts a wandb datatype to its JSON representation.

    converted = val
    typename = util.get_full_typename(val)

    if util.is_pandas_data_frame(val):
        assert step == 'summary', "We don't yet support DataFrames in History. Use `wandb.Table(dataframe=my_dataframe)`"
        return data_frame_to_json(val, run, key, step)
    elif util.is_matplotlib_typename(typename) or util.is_plotly_typename(typename):
        val = Plotly.make_plot_media(val)
    elif isinstance(val, collections.Sequence) and all(isinstance(v, WBValue) for v in val):
        # This check will break down if Image/Audio/... have child classes.
        if len(val) and isinstance(val[0], BatchableMedia) and all(isinstance(v, type(val[0])) for v in val):
            return val[0].seq_to_json(val, run, key, step)
        else:
            # TODO(adrian): Good idea to pass on the same key here? Maybe include
            # the array index?
            # There is a bug here: if this array contains two arrays of the same type of
            # anonymous media objects, their eventual names will collide.
            # This used to happen. The frontend doesn't handle heterogenous arrays
            #raise ValueError(
            #    "Mixed media types in the same list aren't supported")
            return [val_to_json(run, key, v, step=step) for v in val]

    if isinstance(val, WBValue):
        if isinstance(val, Media) and not val.is_bound():
            val.bind_to_run(run, key, step)
        return val.to_json(run)

    return converted


def data_frame_to_json(df, run, key, step):
    """!NODOC Encode a Pandas DataFrame into the JSON/backend format.

    Writes the data to a file and returns a dictionary that we use to represent
    it in `Summary`'s.

    Arguments:
        df (pandas.DataFrame): The DataFrame. Must not have columns named
            "wandb_run_id" or "wandb_data_frame_id". They will be added to the
            DataFrame here.
        run (wandb_run.Run): The Run the DataFrame is associated with. We need
            this because the information we store on the DataFrame is derived
            from the Run it's in.
        key (str): Name of the DataFrame, ie. the summary key path in which it's
            stored. This is for convenience, so people exploring the
            directory tree can have some idea of what is in the Parquet files.
        step: History step or "summary".

    Returns:
        A dict representing the DataFrame that we can store in summaries or
        histories. This is the format:
        {
            '_type': 'data-frame',
                # Magic field that indicates that this object is a data frame as
                # opposed to a normal dictionary or anything else.
            'id': 'asdf',
                # ID for the data frame that is unique to this Run.
            'format': 'parquet',
                # The file format in which the data frame is stored. Currently can
                # only be Parquet.
            'project': 'wfeas',
                # (Current) name of the project that this Run is in. It'd be
                # better to store the project's ID because we know it'll never
                # change but we don't have that here. We store this just in
                # case because we use the project name in identifiers on the
                # back end.
            'path': 'media/data_frames/sdlk.parquet',
                # Path to the Parquet file in the Run directory.
        }
    """
    pandas = util.get_module("pandas")
    fastparquet = util.get_module("fastparquet")
    missing_reqs = []
    if not pandas:
        missing_reqs.append('pandas')
    if not fastparquet:
        missing_reqs.append('fastparquet')
    if len(missing_reqs) > 0:
        raise wandb.Error("Failed to save data frame. Please run 'pip install %s'" % ' '.join(missing_reqs))

    data_frame_id = util.generate_id()

    df = df.copy()  # we don't want to modify the user's DataFrame instance.

    for col_name, series in df.items():
        for i, val in enumerate(series):
            if isinstance(val, WBValue):
                series.iat[i] = six.text_type(json.dumps(val_to_json(run, key, val, step)))

    # We have to call this wandb_run_id because that name is treated specially by
    # our filtering code
    df['wandb_run_id'] = pandas.Series(
        [six.text_type(run.id)] * len(df.index), index=df.index)

    df['wandb_data_frame_id'] = pandas.Series(
        [six.text_type(data_frame_id)] * len(df.index), index=df.index)
    frames_dir = os.path.join(run.dir, DATA_FRAMES_SUBDIR)
    util.mkdir_exists_ok(frames_dir)
    path = os.path.join(frames_dir, '{}-{}.parquet'.format(key, data_frame_id))
    fastparquet.write(path, df)

    return {
        'id': data_frame_id,
        '_type': 'data-frame',
        'format': 'parquet',
        'project': run.project_name(),  # we don't have the project ID here
        'entity': run.entity,
        'run': run.id,
        'path': path,
    }



import pytest
from .utils import git_repo
import os
import glob
import sys
from wandb import docker


@pytest.fixture
def auth_config(mocker):
    mocker.patch('wandb.docker.auth_config.resolve_authconfig',
                 lambda reg: {"Username": "test", "Password": "test"})


@pytest.fixture
def auth_config_funky(mocker):
    mocker.patch('wandb.docker.auth_config.resolve_authconfig',
                 lambda reg: {"username": "test"})


def test_docker_registry_custom(request_mocker, auth_config):
    auth_mock = request_mocker.register_uri('GET', "https://gcr.io/v2/", headers={"Www-Authenticate": 'Bearer realm="https://gcr.io/token",service="gcr.io",scope="repository:foo/bar:pull"'},
                                            content=b'{}', status_code=401)
    token_mock = request_mocker.register_uri(
        'GET', "https://gcr.io/token", content=b'{"token": "test"}')
    mani_mock = request_mocker.register_uri(
        'HEAD', "https://gcr.io/v2/foo/bar/manifests/rad", headers={"Docker-Content-Digest": "sha256:sickdigest"})
    digest = docker.image_id_from_registry("gcr.io/foo/bar:rad")
    assert digest == "gcr.io/foo/bar@sha256:sickdigest"


def test_registry_fucked(request_mocker, auth_config):
    auth_mock = request_mocker.register_uri(
        'GET', "https://crap.io/v2/", content=b'{}', status_code=404)
    mani_mock = request_mocker.register_uri(
        'HEAD', "https://crap.io/v2/foo/bar/manifests/rad", status_code=404)
    digest = docker.image_id_from_registry("crap.io/foo/bar:rad")
    assert digest == None


def test_docker_registry_hub(request_mocker, auth_config_funky):
    request_mocker.register_uri('GET', "https://index.docker.io/v2/", headers={"Www-Authenticate": 'Bearer realm="https://auth.docker.io/token",service="registry.docker.io",scope="repository:samalba/my-app:pull,push'},
                                content=b'{}', status_code=401)
    token_mock = request_mocker.register_uri(
        'GET', "https://auth.docker.io/token", content=b'{"token": "test"}')
    mani_mock = request_mocker.register_uri(
        'HEAD', "https://registry-1.docker.io/v2/foo/bar/manifests/rad", headers={"Docker-Content-Digest": "sha256:sickdigest"})
    docker.image_id_from_registry("foo/bar:rad")



import datetime
import io
import os
import time
import json
import yaml
import glob
import pytest
import threading

import wandb.run_manager
from wandb.apis import internal
import wandb
from .utils import git_repo
from wandb import env
from wandb import wandb_socket
from wandb.wandb_run import Run, RESUME_FNAME
from wandb.run_manager import FileEventHandlerThrottledOverwriteMinWait, FileEventHandlerOverwriteDeferred, FileEventHandlerOverwrite, FileEventHandlerOverwriteOnce
from click.testing import CliRunner


def test_check_update_available_equal(request_mocker, capsys, query_viewer):
    "Test update availability in different cases."
    test_cases = [
        ('0.8.10', '0.8.10', False),
        ('0.8.9', '0.8.10', True),
        ('0.8.11', '0.8.10', False),
        ('1.0.0', '2.0.0', True),
        ('0.4.5', '0.4.5a5', False),
        ('0.4.5', '0.4.3b2', False),
        ('0.4.5', '0.4.6b2', True),
        ('0.4.5.alpha', '0.4.4', False),
        ('0.4.5.alpha', '0.4.5', True),
        ('0.4.5.alpha', '0.4.6', True)
    ]

    for current, latest, is_expected in test_cases:
        with CliRunner().isolated_filesystem():
            query_viewer(request_mocker)
            is_avail = _is_update_avail(
                request_mocker, capsys, current, latest)
            assert is_avail == is_expected, "expected {} compared to {} to yield update availability of {}".format(
                current, latest, is_expected)


def _is_update_avail(request_mocker, capsys, current, latest):
    "Set up the run manager and detect if the upgrade message is printed."
    api = internal.Api(
        load_settings=False,
        retry_timedelta=datetime.timedelta(0, 0, 50))
    api.set_current_run_id(123)
    run = Run()
    run_manager = wandb.run_manager.RunManager(run)

    # Without this mocking, during other tests, the _check_update_available
    # function will throw a "mock not found" error, then silently fail without
    # output (just like it would in a normal network failure).
    response = b'{ "info": { "version": "%s" } }' % bytearray(latest, 'utf-8')
    request_mocker.register_uri('GET', 'https://pypi.org/pypi/wandb/json',
                                content=response, status_code=200)
    run_manager._check_update_available(current)

    captured_out, captured_err = capsys.readouterr()
    print(captured_out, captured_err)
    return "To upgrade, please run:" in captured_err


def test_throttle_file_poller(mocker, run_manager):
    emitter = run_manager.emitter
    assert emitter.timeout == 1
    for i in range(100):
        with open(os.path.join(wandb.run.dir, "file_%i.txt" % i), "w") as f:
            f.write(str(i))
    run_manager.test_shutdown()
    assert emitter.timeout == 2


def test_pip_freeze(mocker, run_manager):
    run_manager._block_file_observer()
    run_manager.init_run()
    # TODO(adrian): I've seen issues with this test when the W&B version
    # installed for the current python differs from the one (eg. from git)
    # that is running this test. Easy fix is to do "pip install -e ."
    reqs = open(os.path.join(wandb.run.dir, "requirements.txt")).read()
    print([r for r in reqs.split("\n") if "pytest" in r])
    wbv = "pytest==%s" % pytest.__version__
    assert wbv in reqs


def test_spell_sync(mocker, loggedin, run_manager, mock_server, local_netrc):
    run_manager._block_file_observer()
    run_manager.init_run(env={"SPELL_RUN_URL": "https://spell.run/test"})
    assert mock_server.ctx['wandb_url'][0]["url"]


def test_custom_file_policy(mocker, run_manager):
    run_manager._block_file_observer()
    run_manager.init_run()
    for i in range(5):
        with open(os.path.join(wandb.run.dir, "ckpt_%i.txt" % i), "w") as f:
            f.write(str(i))
    wandb.save("ckpt*")
    with open(os.path.join(wandb.run.dir, "foo.bar"), "w") as f:
        f.write("bar")

    run_manager.test_shutdown()
    assert isinstance(
        run_manager._file_event_handlers["ckpt_0.txt"], FileEventHandlerThrottledOverwriteMinWait)
    assert isinstance(
        run_manager._file_event_handlers["foo.bar"], FileEventHandlerOverwriteDeferred)
    assert isinstance(
        run_manager._file_event_handlers["wandb-metadata.json"], FileEventHandlerOverwriteOnce)
    assert isinstance(
        run_manager._file_event_handlers["requirements.txt"], FileEventHandlerOverwrite)


def test_custom_file_policy_symlink(mocker, run_manager):
    mod = mocker.MagicMock()
    mocker.patch(
        'wandb.run_manager.FileEventHandlerThrottledOverwriteMinWait.on_modified', mod)
    with open("ckpt_0.txt", "w") as f:
        f.write("joy")
    with open("ckpt_1.txt", "w") as f:
        f.write("joy" * 100)
    wandb.save("ckpt_0.txt")
    with open("ckpt_0.txt", "w") as f:
        f.write("joy" * 100)
    wandb.save("ckpt_1.txt")
    run_manager.test_shutdown()
    assert isinstance(
        run_manager._file_event_handlers["ckpt_0.txt"], FileEventHandlerThrottledOverwriteMinWait)
    assert mod.called


@pytest.mark.skip(reason='batch file archiving removed in favor of more parallelism')
def test_file_pusher_doesnt_archive_if_few(mocker, run_manager, mock_server):
    "Test that only 3 files are uploaded individually."

    # Mock to increase minimum since some extra files are included with all
    # uploads, increasing the number past the default minimum of 6
    from wandb.file_pusher import FilePusher
    mocker.patch.object(FilePusher, 'BATCH_THRESHOLD_SECS', 0.3)
    mocker.patch.object(FilePusher, 'BATCH_MIN_FILES', 10)

    for i in range(2):
        fname = "ckpt_{}.txt".format(i)
        with open(fname, "w") as f:
            f.write("w&b" * 100)
            wandb.save(fname)
    run_manager.test_shutdown()

    filenames = [
        r['variables']['files'][0]
        for r in mock_server.ctx['graphql']
        if 'files' in r['variables']]

    # assert there is no batching
    assert all('.tgz' not in filename for filename in filenames)


@pytest.mark.skip(reason='batch file archiving removed in favor of more parallelism')
def test_file_pusher_archives_multiple(mocker, run_manager, mock_server):
    "Test that 100 files are batched."
    for i in range(10):
        fname = "ckpt_{}.txt".format(i)
        with open(fname, "w") as f:
            f.write("w&b" * 100)
            wandb.save(fname)
    run_manager.test_shutdown()

    req = [r for r in mock_server.ctx['graphql']
           if 'files' in r['variables']][0]

    assert 'query Model' in req['query']
    assert req['variables']['name'] == 'testing'
    assert req['variables']['files'] == ['___batch_archive_1.tgz']


def test_remove_auto_resume(mocker, run_manager):
    resume_path = os.path.join(wandb.wandb_dir(), RESUME_FNAME)
    with open(resume_path, "w") as f:
        f.write("{}")
    run_manager.test_shutdown()
    assert not os.path.exists(resume_path)

def test_code_path_in_config(mocker, git_repo):
    mocker.patch('wandb._get_python_type', lambda: "jupyter")
    os.environ[env.NOTEBOOK_NAME] = "test.ipynb"
    os.environ[env.SAVE_CODE] = "true"
    with open("test.ipynb", "w") as f:
        f.write("{}")
    run = Run()
    run.config.set_run_dir(run.dir)
    run_manager = wandb.run_manager.RunManager(run)
    config_path = os.path.join(run.dir, "config.yaml")
    assert run.config["_wandb"]['code_path'] == "code/test.ipynb"
    print(glob.glob(run.dir + "/*"))
    assert yaml.load(open(config_path))["_wandb"]["value"]["code_path"] == "code/test.ipynb"


def test_sync_etc_multiple_messages(mocker, run_manager):
    mocked_policy = mocker.MagicMock()
    run_manager.update_user_file_policy = mocked_policy
    payload = json.dumps(
        {"save_policy": {"glob": "*.foo", "policy": "end"}}).encode("utf8")
    wandb.run.socket.connection.sendall(payload + b"\0" + payload + b"\0")
    run_manager.test_shutdown()
    assert len(mocked_policy.mock_calls) == 2


def test_init_run_network_down(mocker, caplog):
    with CliRunner().isolated_filesystem():
        mocker.patch("wandb.apis.internal.Api.HTTP_TIMEOUT", 0.5)
        api = internal.Api(
            load_settings=False,
            retry_timedelta=datetime.timedelta(0, 0, 50))
        api.set_current_run_id(123)
        run = Run()
        mocker.patch("wandb.run_manager.RunManager._upsert_run",
                     lambda *args: time.sleep(0.6))
        rm = wandb.run_manager.RunManager(run)
        step = rm.init_run()
        assert step == 0
        assert "Failed to connect" in caplog.text



import numbers
import wandb
from ray import tune

# ray 0.8.1 reorganized ray.tune.util -> ray.tune.utils
try:
    from ray.tune.utils import flatten_dict
except ImportError:
    from ray.tune.util import flatten_dict


class WandbLogger(tune.logger.Logger):
    """Pass WandbLogger to the loggers argument of tune.run

       tune.run("PG", loggers=[WandbLogger], config={
           "monitor": True, "env_config": {
               "wandb": {"project": "my-project-name"}}})
    """

    def _init(self):
        self._config = None
        wandb.init(**self.config.get("env_config", {}).get("wandb", {}))

    def on_result(self, result):
        config = result.get("config")
        if config and self._config is None:
            for k in config.keys():
                # ignore ray/tune 0.9+ callbacks config key
                if k == "callbacks":
                    continue
                if wandb.config.get(k) is None:
                    wandb.config[k] = config[k]
            self._config = config
        tmp = result.copy()
        for k in ["done", "config", "pid", "timestamp"]:
            if k in tmp:
                del tmp[k]
        metrics = {}
        for key, value in flatten_dict(tmp, delimiter="/").items():
            if not isinstance(value, numbers.Number):
                continue
            metrics[key] = value
        wandb.log(metrics)

    def close(self):
        wandb.join()



from __future__ import absolute_import

import sys
import six
import argparse
import copy
import json
import os
import yaml
import importlib
import re

import wandb
from wandb import trigger


_import_hook = None
_run_once = False
_args_argparse = None
_args_system = None
_args_absl = None
_magic_init_seen = False
_magic_config = {}


# PEP302 new import hooks, in python3 we could use importlib
class ImportMetaHook():
    def __init__(self, watch=(), on_import=None):
        self.modules = {}
        self.watch_full = frozenset(())
        self.watch_last = frozenset(())
        self.on_import_full = {}
        self.on_import_last = {}

    def add(self, fullname=None, lastname=None, on_import=None):
        if fullname:
            self.on_import_full[fullname] = on_import
            self.watch_full = frozenset(tuple(self.on_import_full.keys()))
        if lastname:
            self.on_import_last[lastname] = on_import
            self.watch_last = frozenset(tuple(self.on_import_last.keys()))

    def install(self):
        sys.meta_path.insert(0, self)

    def uninstall(self):
        sys.meta_path.remove(self)

    def find_module(self, fullname, path=None):
        if fullname in self.watch_full:
            return self
        lastname = fullname.split('.')[-1]
        if lastname in self.watch_last:
            return self

    def load_module(self, fullname):
        self.uninstall()
        mod = importlib.import_module(fullname)
        self.install()
        self.modules[fullname] = mod
        on_import = self.on_import_full.get(fullname)
        if not on_import:
            lastname = fullname.split('.')[-1]
            on_import = self.on_import_last.get(lastname)
        if on_import:
            on_import(fullname)
        return mod

    def get_modules(self):
        return tuple(self.modules)

    def get_module(self, module):
        return self.modules[module]


class ArgumentException(Exception):
    pass


class SafeArgumentParser(argparse.ArgumentParser):
    def error(self, message):
        raise ArgumentException()


def _merge_dicts(source, destination):
    for key, value in source.items():
        if isinstance(value, dict):
            node = destination.setdefault(key, {})
            _merge_dicts(value, node)
        else:
            destination[key] = value
    return destination


def _dict_from_keyval(k, v, json_parse=True):
    d = ret = {}
    keys = k.split('.')
    for k in keys[:-1]:
        d = d.setdefault(k, {})
    if json_parse:
        try:
            v = json.loads(v.strip('"'))
        except ValueError:
            pass
    d[keys[-1]] = v
    return ret


def _magic_get_config(k, default):
    d = _magic_config
    keys = k.split('.')
    for k in keys[:-1]:
        d = d.get(k, {})
    return d.get(keys[-1], default)


_magic_defaults = {
    'enable': None,
    #'wandb': {
    #    'disable': None,
    #},
    'keras': {
        'fit': {
            'callbacks': {
                'tensorboard': {
                    'enable': True,
                    'duplicate': False,
                    'overwrite': False,
                    'write_graph': None,
                    'histogram_freq': None,
                    'update_freq': None,
                    'write_grads': None,
                    'write_images': None,
                    'batch_size': None,
                    },
                'wandb': {
                    'enable': True,
                    'duplicate': False,
                    'overwrite': False,
                    'log_gradients': None,
                    'log_weights': None,
                    'data_type': "auto",
                    'input_type': None,
                    'output_type': None,
                    'log_evaluation': None,
                    'labels': None,
                    'predictions': None,
                    'save_model': None,
                    'save_weights_only': None,
                    'monitor': None,
                    'mode': None,
                    'verbose': None,
                    },
                'epochs': None,
                'batch_size': None,
                }
            },
        #'compile': {
        #       'optimizer': {
        #           'name': False,
        #           },
        #       'loss': None,
        #    },
        },
    'args': {
        'absl': None,
        'argparse': None,
        'sys': None,
        },
    }
            

def _parse_magic(val):
    # attempt to treat string as a json
    not_set = {}
    if val is None:
        return _magic_defaults, not_set
    if val.startswith("{"):
        try:
            val = json.loads(val)
        except ValueError:
            wandb.termwarn("Unable to parse magic json", repeat=False)
            return _magic_defaults, not_set
        conf = _merge_dicts(_magic_defaults, {})
        return _merge_dicts(val, conf), val
    if os.path.isfile(val):
        try:
            with open(val, 'r') as stream:
                val = yaml.safe_load(stream)
        except IOError as e:
            wandb.termwarn("Unable to read magic config file", repeat=False)
            return _magic_defaults, not_set
        except yaml.YAMLError as e:
            wandb.termwarn("Unable to parse magic yaml file", repeat=False)
            return _magic_defaults, not_set
        conf = _merge_dicts(_magic_defaults, {})
        return _merge_dicts(val, conf), val
    # parse as a list of key value pairs
    if val.find('=') > 0:
        # split on commas but ignore commas inside quotes
        # Using this re allows env variable parsing like:
        # WANDB_MAGIC=key1='"["cat","dog","pizza"]"',key2=true
        items = re.findall(r'(?:[^\s,"]|"(?:\\.|[^"])*")+', val)
        conf_set = {}
        for kv in items:
            kv = kv.split('=')
            if len(kv) != 2:
                wandb.termwarn("Unable to parse magic key value pair", repeat=False)
                continue
            d = _dict_from_keyval(*kv)
            _merge_dicts(d, conf_set)
        conf = _merge_dicts(_magic_defaults, {})
        return _merge_dicts(conf_set, conf), conf_set
    wandb.termwarn("Unable to parse magic parameter", repeat=False)
    return _magic_defaults, not_set
        

def set_entity(value, env=None):
    if env is None:
        env = os.environ


def _fit_wrapper(self, fn, generator=None, *args, **kwargs):
    trigger.call('on_fit')
    keras = sys.modules.get("keras", None)
    tfkeras = sys.modules.get("tensorflow.python.keras", None)
    epochs = kwargs.pop("epochs", None)
    batch_size = kwargs.pop("batch_size", None)

    magic_epochs = _magic_get_config("keras.fit.epochs", None)
    if magic_epochs is not None:
        epochs = magic_epochs
    magic_batch_size = _magic_get_config("keras.fit.batch_size", None)
    if magic_batch_size is not None:
        batch_size = magic_batch_size
    callbacks = kwargs.pop("callbacks", [])

    tb_enabled = _magic_get_config("keras.fit.callbacks.tensorboard.enable", None)
    if tb_enabled:
        k = getattr(self, '_keras_or_tfkeras', None)
        if k:
            tb_duplicate = _magic_get_config("keras.fit.callbacks.tensorboard.duplicate", None)
            tb_overwrite = _magic_get_config("keras.fit.callbacks.tensorboard.overwrite", None)
            tb_present = any([isinstance(cb, k.callbacks.TensorBoard) for cb in callbacks])
            if tb_present and tb_overwrite:
                callbacks = [cb for cb in callbacks if not isinstance(cb, k.callbacks.TensorBoard)]
            if tb_overwrite or tb_duplicate or not tb_present:
                tb_callback_kwargs = {'log_dir': wandb.run.dir}
                cb_args = ('write_graph','histogram_freq', 'update_freq', 'write_grads',
                           'write_images','batch_size')
                for cb_arg in cb_args:
                    v = _magic_get_config("keras.fit.callbacks.tensorboard." + cb_arg, None)
                    if v is not None:
                        tb_callback_kwargs[cb_arg] = v
                tb_callback = k.callbacks.TensorBoard(**tb_callback_kwargs)
                callbacks.append(tb_callback)
    
    wandb_enabled = _magic_get_config("keras.fit.callbacks.wandb.enable", None)
    if wandb_enabled:
        wandb_duplicate = _magic_get_config("keras.fit.callbacks.wandb.duplicate", None)
        wandb_overwrite = _magic_get_config("keras.fit.callbacks.wandb.overwrite", None)
        wandb_present = any([isinstance(cb, wandb.keras.WandbCallback) for cb in callbacks])
        if wandb_present and wandb_overwrite:
            callbacks = [cb for cb in callbacks if not isinstance(cb, wandb.keras.WandbCallback)]
        if wandb_overwrite or wandb_duplicate or not wandb_present:
            wandb_callback_kwargs = {}
            log_gradients = _magic_get_config("keras.fit.callbacks.wandb.log_gradients", None)
            if log_gradients and kwargs.get('x') and kwargs.get('y'):
                wandb_callback_kwargs['log_gradients'] = log_gradients
            cb_args = ("predictions", "log_weights", "data_type", "save_model", "save_weights_only",
                       "monitor", "mode", "verbose", "input_type", "output_type", "log_evaluation",
                       "labels")
            for cb_arg in cb_args:
                v = _magic_get_config("keras.fit.callbacks.wandb." + cb_arg, None)
                if v is not None:
                    wandb_callback_kwargs[cb_arg] = v
            wandb_callback = wandb.keras.WandbCallback(**wandb_callback_kwargs)
            callbacks.append(wandb_callback)

    kwargs["callbacks"] = callbacks
    if epochs is not None:
        kwargs["epochs"] = epochs
    if batch_size is not None:
        kwargs["batch_size"] = batch_size
    if generator:
        return fn(generator, *args, **kwargs)
    return fn(*args, **kwargs)


# NOTE(jhr): need to spell out all useable args so that users who inspect can see args
def _magic_fit(self,
        x=None,
        y=None,
        batch_size=None,
        epochs=1,
        # FIXME: there is more
        #verbose=1,
        #callbacks=None,
        #validation_split=0.,
        #validation_data=None,
        #shuffle=True,
        #class_weight=None,
        #sample_weight=None,
        #initial_epoch=0,
        #steps_per_epoch=None,
        #validation_steps=None,
        #validation_freq=1,
        #max_queue_size=10,
        #workers=1,
        #use_multiprocessing=False,
        *args, **kwargs):
    return _fit_wrapper(self, self._fit, x=x, y=y, batch_size=batch_size, epochs=epochs, *args, **kwargs)


def _magic_fit_generator(self, generator,
                    steps_per_epoch=None,
                    epochs=1,
                    # FIXME: there is more
                    #verbose=1,
                    #verbose=1,
                    #callbacks=None,
                    #validation_data=None,
                    #validation_steps=None,
                    #validation_freq=1,
                    #class_weight=None,
                    #max_queue_size=10,
                    #workers=1,
                    ##use_multiprocessing=False,
                    #shuffle=True,
                    #initial_epoch=0,
                    *args, **kwargs):
    return _fit_wrapper(self, self._fit_generator, generator=generator, steps_per_epoch=steps_per_epoch, epochs=epochs, *args, **kwargs)


def _monkey_keras(keras):
    models = getattr(keras, 'engine', None)
    if not models:
        return
    models.Model._fit = models.Model.fit
    models.Model.fit = _magic_fit
    models.Model._fit_generator = models.Model.fit_generator
    models.Model.fit_generator = _magic_fit_generator
    models.Model._keras_or_tfkeras = keras


def _monkey_tfkeras(tfkeras):
    models = getattr(tfkeras, 'models', None)
    if not models:
        return
    models.Model._fit = models.Model.fit
    models.Model.fit = _magic_fit
    models.Model._fit_generator = models.Model.fit_generator
    models.Model.fit_generator = _magic_fit_generator
    models.Model._keras_or_tfkeras = tfkeras


def _monkey_absl(absl_app):
    def _absl_callback():
        absl_flags = sys.modules.get('absl.flags')
        if not absl_flags:
            return
        _flags = getattr(absl_flags, 'FLAGS', None)
        if not _flags:
            return
        _flags_as_dict = getattr(_flags, 'flag_values_dict', None)
        if not _flags_as_dict:
            return
        _flags_module = getattr(_flags, 'find_module_defining_flag', None)
        if not _flags_module:
            return
        flags_dict = {}
        for f, v in six.iteritems(_flags_as_dict()):
            m = _flags_module(f)
            if not m or m.startswith("absl."):
                continue
            flags_dict[f] = v
        global _args_absl
        _args_absl = flags_dict

    call_after_init = getattr(absl_app, 'call_after_init', None)
    if not call_after_init:
        return
    call_after_init(_absl_callback)


def _on_import_keras(fullname):
    if fullname == 'keras':
        keras = _import_hook.get_module('keras')
        _monkey_keras(keras)
    if fullname == 'tensorflow.python.keras':
        keras = _import_hook.get_module('tensorflow.python.keras')
        _monkey_tfkeras(keras)


def _on_import_absl(fullname):
    if fullname == 'absl.app':
        keras = _import_hook.get_module('absl.app')
        _monkey_absl(keras)


def _process_system_args():
    global _args_system
    # try using argparse
    parser=SafeArgumentParser(add_help=False)
    for num, arg in enumerate(sys.argv):
        try:
            next_arg = sys.argv[num + 1]
        except IndexError:
            next_arg = ''
        if arg.startswith(("-", "--")) and not next_arg.startswith(("-", "--")):
            try:
                parser.add_argument(arg) 
            except ValueError:
                pass
    try:
        parsed, unknown = parser.parse_known_args()
    except ArgumentException:
        pass
    else:
        _args_system = vars(parsed)


def _monkey_argparse():
    argparse._ArgumentParser = argparse.ArgumentParser

    def _install():
        argparse.ArgumentParser = MonitoredArgumentParser

    def _uninstall():
        argparse.ArgumentParser = argparse._ArgumentParser

    def monitored(self, args, unknown=None):
        global _args_argparse
        _args_argparse = copy.deepcopy(vars(args))

    class MonitoredArgumentParser(argparse._ArgumentParser):
        def __init__(self, *args, **kwargs):
            _uninstall()
            super(MonitoredArgumentParser, self).__init__(*args, **kwargs)
            _install()

        def parse_args(self, *args, **kwargs):
            args = super(MonitoredArgumentParser, self).parse_args(*args, **kwargs)
            return args

        def parse_known_args(self, *args, **kwargs):
            args, unknown = super(MonitoredArgumentParser, self).parse_known_args(*args, **kwargs)
            if self._callback:
                self._callback(args, unknown=unknown)
            return args, unknown

    _install()
    argparse.ArgumentParser._callback = monitored
    

def _magic_update_config():
    # if we already have config set, dont add anymore
    if wandb.run and wandb.run.config:
        c = wandb.run.config
        user_config = dict(c.user_items())
        if user_config:
            return
    if _magic_get_config("args.absl", None) is False:
        global _args_absl
        _args_absl = None
    if _magic_get_config("args.argparse", None) is False:
        global _args_argparse
        _args_argparse = None
    if _magic_get_config("args.sys", None) is False:
        global _args_system
        _args_system = None
    # prefer absl, then argparse values, fallback to parsed system args
    args = _args_absl or _args_argparse or _args_system
    if args and wandb.run and wandb.run.config:
        wandb.run.config.update(args)


def _magic_init(**kwargs):
    magic_arg = kwargs.get("magic", None)
    if magic_arg is not None and magic_arg is not False:
        global _magic_init_seen
        if _magic_init_seen and magic_arg is not True:
            wandb.termwarn("wandb.init() magic argument ignored because wandb magic has already been initialized", repeat=False)
        _magic_init_seen = True
    else:
        wandb.termwarn("wandb.init() arguments ignored because wandb magic has already been initialized", repeat=False)


def magic_install(init_args=None):
    global _run_once
    if _run_once:
        return
    _run_once = True

    global _magic_config
    global _import_hook

    # parse config early, before we have wandb.config overrides
    _magic_config, magic_set = _parse_magic(wandb.env.get_magic())

    # we are implicitly enabling magic
    if _magic_config.get('enable') is None:
        _magic_config['enable'] = True
        magic_set['enable'] = True

    # allow early config to disable magic
    if not _magic_config.get('enable'):
        return

    # process system args
    _process_system_args()
    # install argparse wrapper
    in_jupyter_or_ipython = wandb._get_python_type() != "python"
    if not in_jupyter_or_ipython:
        _monkey_argparse()

    # track init calls
    trigger.register('on_init', _magic_init)

    # if wandb.init has already been called, this call is ignored
    init_args = init_args or {}
    init_args['magic'] = True
    wandb.init(**init_args)

    # parse magic from wandb.config (from flattened to dict)
    magic_from_config = {}
    MAGIC_KEY = "wandb_magic"
    for k, v in wandb.config.user_items():
        if not k.startswith(MAGIC_KEY + "."):
            continue
        d = _dict_from_keyval(k, v, json_parse=False)
        _merge_dicts(d, magic_from_config)
    magic_from_config = magic_from_config.get(MAGIC_KEY, {})
    _merge_dicts(magic_from_config, _magic_config)

    # allow late config to disable magic
    if not _magic_config.get('enable'):
        return

    # store magic_set into config
    if magic_set:
        wandb.config._set_wandb('magic', magic_set)
        wandb.config.persist()

    # Monkey patch both tf.keras and keras
    if 'tensorflow.python.keras' in sys.modules:
        _monkey_tfkeras(sys.modules.get('tensorflow.python.keras'))
    if 'keras' in sys.modules:
        _monkey_keras(sys.modules.get('keras'))
    # Always setup import hooks looking for keras or tf.keras
    if not _import_hook:
        _import_hook = ImportMetaHook()
        _import_hook.install()
    _import_hook.add(lastname='keras', on_import=_on_import_keras)

    if 'absl.app' in sys.modules:
        _monkey_absl(sys.modules.get('absl.app'))
    else:
        if not _import_hook:
            _import_hook = ImportMetaHook()
            _import_hook.install()
        _import_hook.add(fullname='absl.app', on_import=_on_import_absl)

    # update wandb.config on fit or program finish
    trigger.register('on_fit', _magic_update_config)
    trigger.register('on_finished', _magic_update_config)



'''W&B plot_feature_importances for catboost

Simple function to send catboost feature importances metrics to the server

Example usage:

catboost_model.fit(...)
.
.
.
wandb.catboost.plot_feature_importances(catboost_model, X_train.columns)
'''

from __future__ import absolute_import, division, print_function, unicode_literals

import numpy as np
import wandb

from .utils import test_fitted, test_types
from warnings import simplefilter

# ignore all future warnings
simplefilter(action='ignore', category=FutureWarning)


def plot_feature_importances(model=None, feature_names=None,
                             title='Feature Importance', max_num_features=50):
    """
    Evaluates & plots the importance of each feature for the classification/regressor tasks.

    Should only be called with a fitted classifer or regressor (otherwise an error is thrown).
    Only works with classifiers that have a feature_importances_ attribute.

    Arguments:
        model (clf): Takes in a fitted classifier or regressor.
        feature_names (list): Names for features. Makes plots easier to read by
                                replacing feature indexes with corresponding names.

    Returns:
        Nothing. To see plots, go to your W&B run page then expand the 'media' tab
              under 'auto visualizations'.

    Example:
        wandb.catboost.plot_feature_importances(model, ['width', 'height, 'length'])
    """
    attributes_to_check = 'feature_importances_'

    if not hasattr(model, attributes_to_check):
        wandb.termwarn(
            "%s attribute not in model. Cannot plot feature importances." % attributes_to_check)
        return

    if (test_types(model=model) and test_fitted(model)):
        feature_names = np.asarray(feature_names)
        importances = model.feature_importances_

        indices = np.argsort(importances)[::-1]

        if feature_names is None:
            feature_names = indices
        else:
            feature_names = np.array(feature_names)[indices]

        max_num_features = min(max_num_features, len(importances))

        feature_names = feature_names[:max_num_features]
        importances = importances[:max_num_features]

        # Draw a stem plot with the influence for each instance
        # format:
        # x = feature_names[:max_num_features]
        # y = importances[indices][:max_num_features]
        def feature_importances_table(feature_names, importances):
            return wandb.visualize(
                'wandb/feature_importances/v1', wandb.Table(
                    columns=['feature_names', 'importances'],
                    data=[
                        [feature_names[i], importances[i]] for i in range(len(feature_names))
                    ]
                ))

        wandb.log({'feature_importances': feature_importances_table(feature_names, importances)})
        print("Go to your dashboard to see the Catboost's Feature Importances graph plotted.")
        return feature_importances_table(feature_names, importances)



import numpy as np
from sklearn.linear_model import ElasticNet
from sklearn.datasets import make_regression, make_classification
from catboost import CatBoostRegressor, CatBoostClassifier

import wandb

from wandb.catboost import plot_feature_importances

# Tests which rely on row history in memory should set `History.keep_rows = True`
from wandb.history import History

History.keep_rows = True


def test_feature_importance_attribute_does_not_exists(wandb_init_run):
    data = np.array([[1, 2], [3, 4]])
    label = np.array([0, 1])
    model = ElasticNet(random_state=42)
    model.fit(data, label)
    dummy_features = []

    result = plot_feature_importances(model, feature_names=dummy_features)

    assert result is None, "Should have NOT returned any result, as feature_importances_ attribute does NOT exist"


def test_feature_importance_attribute_exists_for_catboostregressor(wandb_init_run):
    X, y = make_regression(n_features=2, random_state=42)
    two_features = ['a', 'b']
    model = CatBoostRegressor(random_state=42, early_stopping_rounds=10)
    model = model.fit(X, y)

    result = plot_feature_importances(model, feature_names=two_features)

    assert isinstance(result,
                      wandb.viz.Visualize), "Should have returned a result, as feature_importances_ attribute does exist"


def test_feature_importance_attribute_exists_for_catboostclassifier(wandb_init_run):
    X, y = make_classification(n_samples=100, n_features=10, n_informative=2, n_redundant=5, random_state=42)
    ten_features = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j']
    model = CatBoostClassifier(random_state=42, early_stopping_rounds=10)
    model = model.fit(X, y)

    result = plot_feature_importances(model, feature_names=ten_features)

    assert isinstance(result,
                      wandb.viz.Visualize), "Should have returned a result, as feature_importances_ attribute does exist"



import wandb
from wandb import util
from wandb.plots.utils import test_missing, test_types, encode_labels

def part_of_speech(docs):
        """
        Adds support for spaCy's dependency visualizer which shows
            part-of-speech tags and syntactic dependencies.

        Arguments:
         docs (list, Doc, Span): Document(s) to visualize.

        Returns:
         Nothing. To see plots, go to your W&B run page.

        Example:
         wandb.log({'part_of_speech': wandb.plots.POS(docs=doc)})
        """
        spacy = util.get_module("spacy", required="part_of_speech requires the spacy library, install with `pip install spacy`")
        en_core_web_md = util.get_module("en_core_web_md", required="part_of_speech requires the en_core_web_md library, install with `python -m spacy download en_core_web_md`")
        nlp = en_core_web_md.load()

        if (test_missing(docs=docs)):
            #and test_types(docs=docs)):
            wandb.termlog('Visualizing part of speech.')
            options = {"compact": True, "color": "#1a1c1f",
                "font": "Source Sans Pro", "collapse_punct": True,
                "collapse_phrases": True}
            html = spacy.displacy.render(nlp(str(docs)), style='dep',
                                        minify=True, options=options, page=True)
            return wandb.Html(html)



# -*- coding: utf-8 -*-

"""
Wandb is a library to help track machine learning experiments.

For more information on wandb see https://docs.wandb.com.

The most commonly used functions/objects are:
- wandb.init — initialize a new run at the top of your training script
- wandb.config — track hyperparameters
- wandb.log — log metrics over time within your training loop
- wandb.save — save files in association with your run, like model weights
- wandb.restore — restore the state of your code when you ran a given run

For examples usage, see https://docs.wandb.com/library/example-projects
"""

# Three possible modes:
#     'cli': running from "wandb" command
#     'run': we're a script launched by "wandb run"
#     'dryrun': we're a script not launched by "wandb run"

from __future__ import absolute_import, print_function

__author__ = """Chris Van Pelt"""
__email__ = 'vanpelt@wandb.com'
__version__ = '0.9.3'

import atexit
import click
import io
import json
import logging
import time
import os
import contextlib
import signal
import six
import getpass
import socket
import subprocess
import sys
import traceback
import tempfile
import re
import glob
import threading
import platform
import collections
from six.moves import queue
from six import string_types
from importlib import import_module

from . import env
from . import io_wrap
from .core import *

# These imports need to be below "from .core import *" until we remove
# 'from wandb import __stage_dir__' from api.py etc.
from wandb.apis import InternalApi, PublicApi, CommError
from wandb import wandb_types as types
from wandb import wandb_config
from wandb import wandb_run
from wandb import wandb_socket
from wandb import streaming_log
from wandb import util
from wandb.artifacts import Artifact
from wandb.run_manager import LaunchError, Process
from wandb.data_types import Image
from wandb.data_types import Video
from wandb.data_types import Audio
from wandb.data_types import Table
from wandb.data_types import Html
from wandb.data_types import Object3D
from wandb.data_types import Molecule
from wandb.data_types import Histogram
from wandb.data_types import Graph
from wandb import trigger
from wandb.dataframes import image_categorizer_dataframe
from wandb.dataframes import image_segmentation_dataframe
from wandb.dataframes import image_segmentation_binary_dataframe
from wandb.dataframes import image_segmentation_multiclass_dataframe
from wandb.viz import visualize
from wandb import plots

from wandb import wandb_torch
from wandb.wandb_agent import agent
from wandb.wandb_controller import sweep, controller

from wandb.compat import windows


logger = logging.getLogger(__name__)

# Internal variables
_shutdown_async_log_thread_wait_time = 20

# this global W&B debug log gets re-written by every W&B process
if __stage_dir__ is not None:
    GLOBAL_LOG_FNAME = os.path.abspath(os.path.join(wandb_dir(), 'debug.log'))
else:
    GLOBAL_LOG_FNAME = os.path.join(tempfile.gettempdir(), 'wandb-debug.log')


def _debugger(*args):
    import pdb
    pdb.set_trace()


class Callbacks():
    @property
    def Keras(self):
        termlog(
            "DEPRECATED: wandb.callbacks is deprecated, use `from wandb.keras import WandbCallback`")
        from wandb.keras import WandbCallback
        return WandbCallback


callbacks = Callbacks()


def hook_torch(*args, **kwargs):
    termlog(
        "DEPRECATED: wandb.hook_torch is deprecated, use `wandb.watch`")
    return watch(*args, **kwargs)


_global_watch_idx = 0


def watch(models, criterion=None, log="gradients", log_freq=100, idx=None):
    """
    Hooks into a pytorch model to collect gradients and the topology.  Should be extended
    to accept arbitrary ML models.

    Args:
        models (torch.Module): The model to hook, can be a tuple
        criterion (torch.F, optional): An optional loss value being optimized
        log (str, optional): One of "gradients", "parameters", "all", or None
            if "gradients" watch will only log gradients
            if "parameters" watch will only log parameters
        log_freq (int, optional): log gradients and parameters every N batches
        idx (int, optional): an index to be used when calling wandb.watch on multiple models

    Returns:
        :obj:`wandb.Graph` The graph object that will populate after the first backward pass
    """
    global _global_watch_idx

    # TODO: temporary override for huggingface remove after: https://github.com/huggingface/transformers/pull/4220
    if os.getenv("WANDB_WATCH") == "false":
        return

    if run is None:
        raise ValueError(
            "You must call `wandb.init` before calling watch")

    in_jupyter = _get_python_type() != "python"

    log_parameters = False
    log_gradients = True
    if log == "all":
        log_parameters = True
    elif log == "parameters":
        log_parameters = True
        log_gradients = False
    elif log is None:
        log_gradients = False

    if not isinstance(models, (tuple, list)):
        models = (models,)
    graphs = []
    prefix = ''
    if idx is None:
        idx = _global_watch_idx
    for local_idx, model in enumerate(models):
        global_idx = idx + local_idx
        _global_watch_idx += 1
        if global_idx > 0:
            # TODO: this makes ugly chart names like gradients/graph_1conv1d.bias
            prefix = "graph_%i" % global_idx

        run.history.torch.add_log_hooks_to_pytorch_module(
            model, log_parameters=log_parameters, log_gradients=log_gradients, prefix=prefix, log_freq=log_freq,
            jupyter_run=run if in_jupyter else None)

        graph = wandb_torch.TorchGraph.hook_torch(
            model, criterion, graph_idx=global_idx)
        graphs.append(graph)
        # NOTE: the graph is set in run.summary by hook_torch on the backward pass
    return graphs


def unwatch(models=None):
    """Remove pytorch gradient and parameter hooks.

    Args:
        models (list): Optional list of pytorch models that have had watch called on them
    """
    if models:
        if not isinstance(models, (tuple, list)):
            models = (models,)
        for model in models:
            if not hasattr(model, "_wandb_hook_names"):
                termwarn("%s model has not been watched" % model)
            else:
                for name in model._wandb_hook_names:
                    run.history.torch.unhook(name)
    else:
        run.history.torch.unhook_all()


class ExitHooks(object):
    def __init__(self):
        self.exit_code = 0
        self.exception = None

    def hook(self):
        self._orig_exit = sys.exit
        sys.exit = self.exit
        sys.excepthook = self.exc_handler

    def exit(self, code=0):
        orig_code = code
        if code is None:
            code = 0
        elif not isinstance(code, int):
            code = 1
        self.exit_code = code
        self._orig_exit(orig_code)

    def was_ctrl_c(self):
        return isinstance(self.exception, KeyboardInterrupt)

    def exc_handler(self, exc_type, exc, *tb):
        self.exit_code = 1
        self.exception = exc
        if issubclass(exc_type, Error):
            termerror(str(exc))

        if self.was_ctrl_c():
            self.exit_code = 255

        traceback.print_exception(exc_type, exc, *tb)


def _init_headless(run, cloud=True):
    global join
    global _user_process_finished_called

    program = util.get_program()
    if program:
        os.environ[env.PROGRAM] = os.getenv(env.PROGRAM) or program

    environ = dict(os.environ)
    run.set_environment(environ)

    server = wandb_socket.Server()
    run.socket = server
    hooks = ExitHooks()
    hooks.hook()

    if platform.system() == "Windows":
        try:
            import win32api
            # Make sure we are not ignoring CTRL_C_EVENT
            # https://docs.microsoft.com/en-us/windows/console/setconsolectrlhandler
            # https://stackoverflow.com/questions/1364173/stopping-python-using-ctrlc
            win32api.SetConsoleCtrlHandler(None, False)
        except ImportError:
            termerror("Install the win32api library with `pip install pypiwin32`")

        # PTYs don't work in windows so we create these unused pipes and
        # mirror stdout to run.dir/output.log.  There should be a way to make
        # pipes work, but I haven't figured it out.  See links in compat/windows
        stdout_master_fd, stdout_slave_fd = os.pipe()
        stderr_master_fd, stderr_slave_fd = os.pipe()
    else:
        stdout_master_fd, stdout_slave_fd = io_wrap.wandb_pty(resize=False)
        stderr_master_fd, stderr_slave_fd = io_wrap.wandb_pty(resize=False)

    headless_args = {
        'command': 'headless',
        'pid': os.getpid(),
        'stdout_master_fd': stdout_master_fd,
        'stderr_master_fd': stderr_master_fd,
        'cloud': cloud,
        'port': server.port
    }
    internal_cli_path = os.path.join(
        os.path.dirname(__file__), 'internal_cli.py')

    if six.PY2 or platform.system() == "Windows":
        # TODO(adrian): close_fds=False is bad for security. we set
        # it so we can pass the PTY FDs to the wandb process. We
        # should use subprocess32, which has pass_fds.
        popen_kwargs = {'close_fds': False}
    else:
        popen_kwargs = {'pass_fds': [stdout_master_fd, stderr_master_fd]}

    # TODO(adrian): ensure we use *exactly* the same python interpreter
    # TODO(adrian): make wandb the foreground process so we don't give
    # up terminal control until syncing is finished.
    # https://stackoverflow.com/questions/30476971/is-the-child-process-in-foreground-or-background-on-fork-in-c
    wandb_process = subprocess.Popen([sys.executable, internal_cli_path, json.dumps(
        headless_args)], env=environ, **popen_kwargs)
    termlog('Tracking run with wandb version {}'.format(
        __version__))
    os.close(stdout_master_fd)
    os.close(stderr_master_fd)
    # Listen on the socket waiting for the wandb process to be ready
    try:
        success, _ = server.listen(30)
    except KeyboardInterrupt:
        success = False
    else:
        if not success:
            termerror('W&B process (PID {}) did not respond'.format(
                wandb_process.pid))
    if not success:
        wandb_process.kill()
        for _ in range(20):
            time.sleep(0.1)
            if wandb_process.poll() is not None:
                break
        if wandb_process.poll() is None:
            termerror('Failed to kill wandb process, PID {}'.format(
                wandb_process.pid))
        # TODO attempt to upload a debug log
        path = GLOBAL_LOG_FNAME.replace(os.getcwd()+os.sep, "")
        raise LaunchError(
            "W&B process failed to launch, see: {}".format(path))

    if platform.system() == "Windows":
        output = open(os.path.join(run.dir, "output.log"), "wb")
        stdout_redirector = io_wrap.WindowsRedirector(sys.stdout, output)
        stderr_redirector = io_wrap.WindowsRedirector(sys.stderr, output)
    else:
        stdout_slave = os.fdopen(stdout_slave_fd, 'wb')
        stderr_slave = os.fdopen(stderr_slave_fd, 'wb')
        try:
            stdout_redirector = io_wrap.FileRedirector(sys.stdout, stdout_slave)
            stderr_redirector = io_wrap.FileRedirector(sys.stderr, stderr_slave)
        except (ValueError, AttributeError):
            # stdout / err aren't files
            output = open(os.path.join(run.dir, "output.log"), "wb")
            stdout_redirector = io_wrap.WindowsRedirector(sys.stdout, output)
            stderr_redirector = io_wrap.WindowsRedirector(sys.stderr, output)

    # TODO(adrian): we should register this right after starting the wandb process to
    # make sure we shut down the W&B process eg. if there's an exception in the code
    # above
    atexit.register(_user_process_finished, server, hooks,
                    wandb_process, stdout_redirector, stderr_redirector)

    def _wandb_join(exit_code=None):
        global _global_run_stack
        shutdown_async_log_thread()
        run.close_files()
        if exit_code is not None:
            hooks.exit_code = exit_code
        _user_process_finished(server, hooks,
                               wandb_process, stdout_redirector, stderr_redirector)
        if len(_global_run_stack) > 0:
            _global_run_stack.pop()
    join = _wandb_join
    _user_process_finished_called = False

    # redirect output last of all so we don't miss out on error messages
    stdout_redirector.redirect()
    if not env.is_debug():
        stderr_redirector.redirect()


def load_ipython_extension(ipython):
    pass


def login(anonymous=None, key=None):
    """Ensure this machine is logged in

       You can manually specify a key, but this method is intended to prompt for user input.

       anonymous can be "never", "must", or "allow".  If set to "must" we'll always login anonymously,
       if set to "allow" we'll only create an anonymous user if the user isn't already logged in.

       Returns:
            True if login was successful
            False on failure
    """
    # This ensures we have a global api object
    ensure_configured()
    if anonymous:
        os.environ[env.ANONYMOUS] = anonymous
    anonymous = anonymous or "never"
    in_jupyter = _get_python_type() != "python"
    if key:
        termwarn("If you're specifying your api key in code, ensure this code is not shared publically.\nConsider setting the WANDB_API_KEY environment variable, or running `wandb login` from the command line.")
        if in_jupyter:
            termwarn("Calling wandb.login() without arguments from jupyter should prompt you for an api key.")
        util.set_api_key(api, key)
    elif api.api_key and anonymous != "must":
        key = api.api_key
    elif in_jupyter:
        os.environ[env.JUPYTER] = "true"
        # Don't return key to ensure it's not displayed in the notebook.
        key = _jupyter_login(api=api)
    else:
        key = util.prompt_api_key(api)
    return True if key else False


def _jupyter_login(force=True, api=None):
    """Attempt to login from a jupyter environment

    If force=False, we'll only attempt to auto-login, otherwise we'll prompt the user
    """
    def get_api_key_from_browser(signup=False):
        key, anonymous = None, False
        if 'google.colab' in sys.modules:
            key = jupyter.attempt_colab_login(api.app_url)
        elif 'databricks_cli' in sys.modules and 'dbutils' in sys.modules:
            # Databricks does not seem to support getpass() so we need to fail
            # early and prompt the user to configure the key manually for now.
            termerror(
                "Databricks requires api_key to be configured manually, instructions at: http://docs.wandb.com/integrations/databricks")
            raise LaunchError("Databricks integration requires api_key to be configured.")
        # For jupyter we default to not allowing anonymous
        if not key and os.environ.get(env.ANONYMOUS, "never") != "never":
            key = api.create_anonymous_api_key()
            anonymous = True
        if not key and force:
            try:
                termerror("Not authenticated.  Copy a key from https://app.wandb.ai/authorize")
                key = getpass.getpass("API Key: ").strip()
            except NotImplementedError:
                termerror(
                    "Can't accept input in this environment, you should set WANDB_API_KEY or call wandb.login(key='YOUR_API_KEY')")
        return key, anonymous

    api = api or (run.api if run else None)
    if not api:
        raise LaunchError("Internal error: api required for jupyter login")
    return util.prompt_api_key(api, browser_callback=get_api_key_from_browser)


def _init_jupyter(run):
    """Asks for user input to configure the machine if it isn't already and creates a new run.
    Log pushing and system stats don't start until `wandb.log()` is first called.
    """
    from wandb import jupyter
    from IPython.core.display import display, HTML

    # TODO: Should we log to jupyter?
    # global logging had to be disabled because it set the level to debug
    # I also disabled run logging because we're rairly using it.
    # try_to_set_up_global_logging()
    # run.enable_logging()
    os.environ[env.JUPYTER] = "true"

    if not run.api.api_key:
        # Fetches or prompts the users for an API key. Or if anonymode enabled, uses anonymous API key
        key = _jupyter_login()
        # Ensure our api client picks up the new key
        if key:
            run.api.reauth()
            # Now that we have an api key, let's load the viewer
            run._load_viewer()
        else:
            run.mode = "dryrun"
            display(HTML('''
                <b>Could not authenticate.</b><br/>
            '''))
    run.resume = "allow"
    if run.mode == "dryrun":
        display(HTML('''
            Using <a href="https://wandb.com" target="_blank">Weights & Biases</a> in dryrun mode. Not logging results to the cloud.<br/>
            Call wandb.login() to authenticate this machine.<br/>
        '''.format(run.api.app_url)))
    else:
        displayed = False
        try:
            sweep_url = run.get_sweep_url()
            sweep_line = 'Sweep page: <a href="{}" target="_blank">{}</a><br/>\n'.format(
                sweep_url, sweep_url) if sweep_url else ""
            docs_html = '<a href="https://docs.wandb.com/integrations/jupyter.html" target="_blank">(Documentation)</a>'
            display(HTML('''
                Logging results to <a href="https://wandb.com" target="_blank">Weights & Biases</a> {}.<br/>
                Project page: <a href="{}" target="_blank">{}</a><br/>
                {}Run page: <a href="{}" target="_blank">{}</a><br/>
            '''.format(docs_html, run.get_project_url(), run.get_project_url(), sweep_line, run.get_url(), run.get_url() )))
            displayed = True
            run.save()
        except (CommError, ValueError) as e:
            if not displayed:
                display(HTML('''
                    Logging results to <a href="https://wandb.com" target="_blank">Weights & Biases</a>.<br/>
                    Couldn't load entity due to error: {}
                '''.format(e.message)))
            else:
                termerror(str(e))

    run.set_environment()
    run._init_jupyter_agent()
    ipython = get_ipython()
    ipython.register_magics(jupyter.WandBMagics)

    # Monkey patch ipython publish to capture displayed outputs
    if not hasattr(ipython.display_pub, "_orig_publish"):
        ipython.display_pub._orig_publish = ipython.display_pub.publish
    def publish(data, metadata=None, **kwargs):
        ipython.display_pub._orig_publish(data, metadata=metadata, **kwargs)
        run._jupyter_agent.save_display(ipython.execution_count , {'data':data, 'metadata':metadata})
    ipython.display_pub.publish = publish

    # Cell start
    def reset_start():
        """Reset START_TIME to when the cell starts"""
        global START_TIME
        START_TIME = time.time()
    if hasattr(ipython.events, "_orig_pre_run"):
        ipython.events.unregister("pre_run_cell", ipython.events._orig_pre_run)
    ipython.events._orig_pre_run = reset_start
    ipython.events.register("pre_run_cell", reset_start)

    # Cell shutdown
    def cleanup():
        # shutdown async logger because _user_process_finished isn't called in jupyter
        shutdown_async_log_thread()
        run._stop_jupyter_agent()
    if hasattr(ipython.events, "_orig_post_run"):
        ipython.events.unregister("post_run_cell", ipython.events._orig_post_run)
    ipython.events._orig_post_run = cleanup
    ipython.events.register('post_run_cell', cleanup)


_user_process_finished_called = False


def _user_process_finished(server, hooks, wandb_process, stdout_redirector, stderr_redirector):
    global _user_process_finished_called
    if _user_process_finished_called:
        return
    _user_process_finished_called = True
    trigger.call('on_finished')
    if run:
        run.close_files()

    stdout_redirector.restore()
    if not env.is_debug():
        stderr_redirector.restore()

    termlog()
    termlog("Waiting for W&B process to finish, PID {}".format(wandb_process.pid))
    server.done(hooks.exit_code)
    try:
        while wandb_process.poll() is None:
            time.sleep(0.1)
    except KeyboardInterrupt:
        termlog('Sending ctrl-c to W&B process, PID {}. Press ctrl-c again to kill it.'.format(wandb_process.pid))

    try:
        while wandb_process.poll() is None:
            time.sleep(0.1)
    except KeyboardInterrupt:
        if wandb_process.poll() is None:
            termlog('Killing W&B process, PID {}'.format(wandb_process.pid))
            wandb_process.kill()


# Will be set to the run object for the current run, as returned by
# wandb.init(). We may want to get rid of this, but WandbCallback
# relies on it, and it improves the API a bit (user doesn't have to
# pass the run into WandbCallback).  run is None instead of a PreInitObject
# as many places in the code check this.
run = None
config = util.PreInitObject("wandb.config")  # config object shared with the global run
summary = util.PreInitObject("wandb.summary")  # summary object shared with the global run
Api = PublicApi
# Stores what modules have been patched
patched = {
    "tensorboard": [],
    "keras": [],
    "gym": []
}
_saved_files = set()
_global_run_stack = []


def join(exit_code=None):
    """Marks a run as finished"""
    shutdown_async_log_thread()
    if run:
        run.close_files()
    if len(_global_run_stack) > 0:
        _global_run_stack.pop()

def use_artifact(artifact_or_name, type=None, aliases=None):
    """ Declare an artifact as an input to a run, call `download` or `file` on \
        the returned object to get the contents locally.

        Args:
            artifact_or_name (str or Artifact): An artifact name. May be prefixed with entity/project. Valid names
                can be in the following forms:
                    name:version
                    name:alias
                    digest
                You can also pass an Artifact object created by calling `wandb.Artifact`
            type (str, optional): The type of artifact to use.
            aliases (list, optional): Aliases to apply to this artifact
        Returns:
            A :obj:`Artifact` object.
    """
    if run is None:
        raise ValueError(
            "You must call `wandb.init` before calling use_artifact")
    return run.use_artifact(artifact_or_name, type, aliases)


def log_artifact(artifact_or_path, name=None, type=None, aliases=['latest']):
    """ Declare an artifact as output of a run.

        Args:
            artifact_or_path (str or Artifact): A path to the contents of this artifact,
                can be in the following forms:
                    /local/directory
                    /local/directory/file.txt
                    s3://bucket/path
                You can also pass an Artifact object created by calling `wandb.Artifact`
            name (str, optional): An artifact name. May be prefixed with entity/project. Valid names
                can be in the following forms:
                    name:version
                    name:alias
                    digest
                this will default to the basename of the path prepended with the current run id if not specified
            type (str): The type of artifact to log, examples include "dataset", "model"
            aliases (list, optional): Aliases to apply to this artifact, defaults to ["latest"]
        Returns:
            A :obj:`Artifact` object.
    """
    if run is None:
        raise ValueError(
            "You must call `wandb.init` before calling log_artifact")
    return run.log_artifact(artifact_or_path, name, type, aliases)


def save(glob_str, base_path=None, policy="live"):
    """ Ensure all files matching *glob_str* are synced to wandb with the policy specified.

    Args:
        base_path (string): the base path to run the glob relative to
        policy (string): on of "live" or "end"
            live: upload the file as it changes, overwriting the previous version
            end: only upload file when the run ends
    """
    global _saved_files
    if run is None:
        raise ValueError(
            "You must call `wandb.init` before calling save")
    if policy not in ("live", "end"):
        raise ValueError(
            'Only "live" and "end" policies are currently supported.')
    if isinstance(glob_str, bytes):
        glob_str = glob_str.decode('utf-8')
    if not isinstance(glob_str, string_types):
        raise ValueError("Must call wandb.save(glob_str) with glob_str a str")

    if base_path is None:
        base_path = os.path.dirname(glob_str)
    wandb_glob_str = os.path.relpath(glob_str, base_path)
    if "../" in wandb_glob_str:
        raise ValueError(
            "globs can't walk above base_path")
    if (glob_str, base_path, policy) in _saved_files:
        return []
    if glob_str.startswith("gs://") or glob_str.startswith("s3://"):
        termlog(
            "%s is a cloud storage url, can't save file to wandb." % glob_str)
        return []
    run.send_message(
        {"save_policy": {"glob": wandb_glob_str, "policy": policy}})
    files = []
    for path in glob.glob(glob_str):
        file_name = os.path.relpath(path, base_path)
        abs_path = os.path.abspath(path)
        wandb_path = os.path.join(run.dir, file_name)
        util.mkdir_exists_ok(os.path.dirname(wandb_path))
        # We overwrite existing symlinks because namespaces can change in Tensorboard
        if os.path.islink(wandb_path) and abs_path != os.readlink(wandb_path):
            os.remove(wandb_path)
            os.symlink(abs_path, wandb_path)
        elif not os.path.exists(wandb_path):
            os.symlink(abs_path, wandb_path)
        files.append(wandb_path)
    _saved_files.add((glob_str, base_path, policy))
    return files


def restore(name, run_path=None, replace=False, root=None):
    """ Downloads the specified file from cloud storage into the current run directory
    if it doesn exist.

    Args:
        name: the name of the file
        run_path: optional path to a different run to pull files from
        replace: whether to download the file even if it already exists locally
        root: the directory to download the file to.  Defaults to the current
            directory or the run directory if wandb.init was called.

    Returns:
        None if it can't find the file, otherwise a file object open for reading

    Raises:
        wandb.CommError if it can't find the run
    """

    if run_path is None and run is None:
        raise ValueError(
            "You must call `wandb.init` before calling restore or specify a run_path")
    api = Api()
    api_run = api.run(run_path or run.path)
    if root is None:
        root = run.dir if run else '.'
    path = os.path.join(root, name)
    if os.path.exists(path) and replace == False:
        return open(path, "r")
    files = api_run.files([name])
    if len(files) == 0:
        return None
    return files[0].download(root=root, replace=True)


_tunnel_process = None


def tunnel(host, port):
    """Simple helper to open a tunnel.  Returns a public HTTPS url or None"""
    global _tunnel_process
    if _tunnel_process:
        _tunnel_process.kill()
        _tunnel_process = None
    process = subprocess.Popen("ssh -o StrictHostKeyChecking=no -o ServerAliveInterval=60 -R 80:{}:{} serveo.net".format(
        host, port), shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    while process.returncode is None:
        for line in process.stdout:
            match = re.match(r".+(https.+)$", line.decode("utf-8").strip())
            if match:
                _tunnel_process = process
                return match.group(1)
        # set returncode if the process has exited
        process.poll()
        time.sleep(1)
    return None


def monitor(options={}):
    """Starts syncing with W&B if you're in Jupyter.  Displays your W&B charts live in a Jupyter notebook.
    It's currently a context manager for legacy reasons.
    """
    try:
        from IPython.display import display
    except ImportError:
        def display(stuff): return None

    class Monitor():
        def __init__(self, options={}):
            if os.getenv(env.JUPYTER):
                display(jupyter.Run())
            else:
                self.rm = False
                termerror(
                    "wandb.monitor is only functional in Jupyter notebooks")

        def __enter__(self):
            termlog(
                "DEPRECATED: with wandb.monitor(): is deprecated, add %%wandb to the beginning of a cell to see live results.")
            pass

        def __exit__(self, *args):
            pass

    return Monitor(options)


_async_log_queue = queue.Queue()
_async_log_thread_shutdown_event = threading.Event()
_async_log_thread_complete_event = threading.Event()
_async_log_thread = None


def _async_log_thread_target():
    """Consumes async logs from our _async_log_queue and actually logs them"""
    global _async_log_thread
    shutdown_requested = False
    while not shutdown_requested:
        try:
            kwargs = _async_log_queue.get(block=True, timeout=1)
            log(**kwargs)
        except queue.Empty:
            shutdown_requested = _async_log_thread_shutdown_event.wait(1) and _async_log_queue.empty()
    _async_log_thread_complete_event.set()
    _async_log_thread = None


def _ensure_async_log_thread_started():
    """Ensures our log consuming thread is started"""
    global _async_log_thread, _async_log_thread_shutdown_event, _async_log_thread_complete_event

    if _async_log_thread is None:
        _async_log_thread_shutdown_event = threading.Event()
        _async_log_thread_complete_event = threading.Event()
        _async_log_thread = threading.Thread(target=_async_log_thread_target)
        _async_log_thread.daemon = True
        _async_log_thread.start()


def shutdown_async_log_thread():
    """Shuts down our async logging thread"""
    if _async_log_thread:
        _async_log_thread_shutdown_event.set()
        res = _async_log_thread_complete_event.wait(_shutdown_async_log_thread_wait_time)  # TODO: possible race here
        if res is False:
            termwarn('async log queue not empty after %d seconds, some log statements will be dropped' % (
                _shutdown_async_log_thread_wait_time))
            # FIXME: it is worse than this, likely the program will crash because files will be closed
        # FIXME: py 2.7 will return None here so we dont know if we dropped data


def log(row=None, commit=None, step=None, sync=True, *args, **kwargs):
    """Log a dict to the global run's history.

    wandb.log can be used to log everything from scalars to histograms, media and matplotlib plots.

    The most basic usage is wandb.log({'train-loss': 0.5, 'accuracy': 0.9}). This will save a history
        row associated with the run with train-loss=0.5 and accuracy=0.9. The history values can be plotted
        on app.wandb.ai or on a local server. The history values can also be downloaded throught the wandb API.

    Logging a value will update the summary values for any metrics logged. The summary values will
        appear in the run table at app.wandb.ai or a local server. If a summary value is manually set
        with for example wandb.run.summary["accuracy"] = 0.9 wandb.log will no longer automatically update
        the run's accuracy.

    Logging values don't have to be scalars. Logging any wandb object is supported. For example
        wandb.log({"example": wandb.Image("myimage.jpg")}) will log an example image which will be
        displayed nicely in the wandb UI. See https://docs.wandb.com/library/reference/data_types
        for all of the different supported types.

    Logging nested metrics is encouraged and is supported in the wandb API, so you could log multiple accuracy
        values with wandb.log({'dataset-1-metrics': {'accuracy': 0.9, 'loss': 0.3} ,'dataset-1-metrics': {'accuracy': 0.8, 'loss': 0.2}})
        and the metrics will be organized in the wandb UI.

    Wandb keeps track of a global step so logging related metrics together is encouraged, so by default each
        time wandb.log is called a global step is incremented. If it's inconvenient to log related metrics together
        calling wandb.log({'train-loss': 0.5, commit=False}) and then wandb.log({'accuracy': 0.9}) is equivalent
        to calling wandb.log({'train-loss': 0.5, 'accuracy': 0.9})

    Wandb.log is not intended to be called more than a few times per second. If you want to log more frequently
        than that it's better to aggregate the data on the client side or you may get degraded performance.

    Args:
        row (dict, optional): A dict of serializable python objects i.e str: ints, floats, Tensors, dicts, or wandb.data_types
        commit (boolean, optional): Save the metrics dict to the wandb server and increment the step.
            If false wandb.log just updates the current metrics dict with the row argument and metrics won't
            be saved until wandb.log is called with commit=True.
        step (integer, optional): The global step in processing. This persists any non-committed earlier steps but defaults
            to not committing the specified step.
        sync (boolean, True): If set to False, process calls to log in a seperate thread.

    Examples:
        Basic usage
        ```
        wandb.log({'accuracy': 0.9, 'epoch': 5})
        ```

        Incremental logging
        ```
        wandb.log({'loss': 0.2}, commit=False)
        # Somewhere else when I'm ready to report this step:
        wandb.log({'accuracy': 0.8})
        ```

        Histogram
        ```
        wandb.log({"gradients": wandb.Histogram(numpy_array_or_sequence)})
        ```

        Image
        ```
        wandb.log({"examples": [wandb.Image(numpy_array_or_pil, caption="Label")]})
        ```

        Video
        ```
        wandb.log({"video": wandb.Video(numpy_array_or_path_to_video, fps=4, format="gif")})
        ```

        Matplotlib Plot
        ```
        wandb.log({"chart": plt})
        ```

        PR Curve
        ```
        wandb.log({'pr': wandb.plots.precision_recall(y_test, y_probas, labels)})
        ```

        3D Object
        ```
        wandb.log({"generated_samples":
           [wandb.Object3D(open("sample.obj")),
            wandb.Object3D(open("sample.gltf")),
            wandb.Object3D(open("sample.glb"))]})
        ```

        For more examples, see https://docs.wandb.com/library/log

    Raises:
        ValueError - if called before wandb.init

    """

    if run is None:
        raise ValueError(
            "You must call `wandb.init` in the same process before calling log")

    run.log(row, commit, step, sync, *args, **kwargs)


def ensure_configured():
    global GLOBAL_LOG_FNAME, api
    # We re-initialize here for tests
    api = InternalApi()
    GLOBAL_LOG_FNAME = os.path.abspath(os.path.join(wandb_dir(), 'debug.log'))


def uninit(only_patches=False):
    """Undo the effects of init(). Useful for testing.
    """
    global run, config, summary, patched, _saved_files
    if not only_patches:
        run = None
        config = util.PreInitObject("wandb.config")
        summary = util.PreInitObject("wandb.summary")
        _saved_files = set()
    # UNDO patches
    for mod in patched["tensorboard"]:
        module = import_module(mod[0])
        parts = mod[1].split(".")
        if len(parts) > 1:
            module = getattr(module, parts[0])
            mod[1] = parts[1]
        setattr(module, mod[1], getattr(module, "orig_"+mod[1]))
    patched["tensorboard"] = []


def reset_env(exclude=[]):
    """Remove environment variables, used in Jupyter notebooks"""
    if os.getenv(env.INITED):
        wandb_keys = [key for key in os.environ.keys() if key.startswith(
            'WANDB_') and key not in exclude]
        for key in wandb_keys:
            del os.environ[key]
        return True
    else:
        return False


def try_to_set_up_global_logging():
    """Try to set up global W&B debug log that gets re-written by every W&B process.

    It may fail (and return False) eg. if the current directory isn't user-writable
    """
    root = logging.getLogger("wandb")
    root.setLevel(logging.DEBUG)
    formatter = logging.Formatter(
        '%(asctime)s %(levelname)-7s %(threadName)-10s:%(process)d [%(filename)s:%(funcName)s():%(lineno)s] %(message)s')

    if env.is_debug():
        handler = logging.StreamHandler()
        handler.setLevel(logging.DEBUG)
        handler.setFormatter(formatter)

        root.addHandler(handler)

    try:
        handler = logging.FileHandler(GLOBAL_LOG_FNAME, mode='w')
        handler.setLevel(logging.DEBUG)
        handler.setFormatter(formatter)

        root.addHandler(handler)
    except IOError as e:  # eg. in case wandb directory isn't writable
        termerror('Failed to set up logging: {}'.format(e))
        return False

    return True


def _get_python_type():
    try:
        if 'terminal' in get_ipython().__module__:
            return 'ipython'
        else:
            return 'jupyter'
    except (NameError, AttributeError):
        return "python"


def sagemaker_auth(overrides={}, path="."):
    """ Write a secrets.env file with the W&B ApiKey and any additional secrets passed.

        Args:
            overrides (dict, optional): Additional environment variables to write to secrets.env
            path (str, optional): The path to write the secrets file.
    """

    api_key = overrides.get(env.API_KEY, Api().api_key)
    if api_key is None:
        raise ValueError(
            "Can't find W&B ApiKey, set the WANDB_API_KEY env variable or run `wandb login`")
    overrides[env.API_KEY] = api_key
    with open(os.path.join(path, "secrets.env"), "w") as file:
        for k, v in six.iteritems(overrides):
            file.write("{}={}\n".format(k, v))


def init(job_type=None, dir=None, config=None, project=None, entity=None, reinit=None, tags=None,
         group=None, allow_val_change=False, resume=False, force=False, tensorboard=False,
         sync_tensorboard=False, monitor_gym=False, name=None, notes=None, id=None, magic=None,
         anonymous=None, config_exclude_keys=None, config_include_keys=None, save_code=None):
    """Initialize W&B

    If called from within Jupyter, initializes a new run and waits for a call to
    `wandb.log` to begin pushing metrics.  Otherwise, spawns a new process
    to communicate with W&B.

    Args:
        job_type (str, optional): The type of job running, defaults to 'train'
        config (dict, argparse, or tf.FLAGS, optional): The hyper parameters to store with the run
        config_exclude_keys (list, optional): string keys to exclude storing in W&B when specifying config
        config_include_keys (list, optional): string keys to include storing in W&B when specifying config
        project (str, optional): The project to push metrics to
        entity (str, optional): The entity to push metrics to
        dir (str, optional): An absolute path to a directory where metadata will be stored
        group (str, optional): A unique string shared by all runs in a given group
        tags (list, optional): A list of tags to apply to the run
        id (str, optional): A globally unique (per project) identifier for the run
        name (str, optional): A display name which does not have to be unique
        notes (str, optional): A multiline string associated with the run
        reinit (bool, optional): Allow multiple calls to init in the same process
        resume (bool, str, optional): Automatically resume this run if run from the same machine,
            you can also pass a unique run_id
        sync_tensorboard (bool, optional): Synchronize wandb logs to tensorboard or tensorboardX
        save_code (bool, optional): Save the entrypoint or jupyter session history source code.
        force (bool, optional): Force authentication with wandb, defaults to False
        magic (bool, dict, or str, optional): magic configuration as bool, dict, json string,
            yaml filename
        anonymous (str, optional): Can be "allow", "must", or "never". Controls whether anonymous logging is allowed.
            Defaults to never.

    Returns:
        A wandb.run object for metric and config logging.
    """
    init_args = locals()
    trigger.call('on_init', **init_args)
    global run
    global __stage_dir__
    global _global_watch_idx

    # TODO: temporary override for huggingface remove after: https://github.com/huggingface/transformers/pull/4220
    if os.getenv("WANDB_DISABLED") == "true":
        return None
    elif wandb_config.huggingface_version() is not None:
        if InternalApi().api_key is None:
            termwarn("W&B installed but not logged in.  Run `wandb login` or set the WANDB_API_KEY env variable.")
            return None

    # We allow re-initialization when we're in Jupyter or explicity opt-in to it.
    in_jupyter = _get_python_type() != "python"
    if reinit or (in_jupyter and reinit != False):
        # Reset global state for pytorch watch and tensorboard
        _global_watch_idx = 0
        if len(patched["tensorboard"]) > 0:
            util.get_module("wandb.tensorboard").reset_state()
        reset_env(exclude=env.immutable_keys())
        if len(_global_run_stack) > 0:
            if len(_global_run_stack) > 1:
                termwarn("If you want to track multiple runs concurrently in wandb you should use multi-processing not threads")
            join()
        run = None

    # TODO: deprecate tensorboard
    if tensorboard or sync_tensorboard and len(patched["tensorboard"]) == 0:
        util.get_module("wandb.tensorboard").patch()
    if monitor_gym and len(patched["gym"]) == 0:
        util.get_module("wandb.gym").monitor()

    sagemaker_config = util.parse_sm_config()
    tf_config = util.parse_tfjob_config()
    if group == None:
        group = os.getenv(env.RUN_GROUP)
    if job_type == None:
        job_type = os.getenv(env.JOB_TYPE)
    if sagemaker_config:
        # Set run_id and potentially grouping if we're in SageMaker
        run_id = os.getenv('TRAINING_JOB_NAME')
        if run_id:
            os.environ[env.RUN_ID] = '-'.join([
                run_id,
                os.getenv('CURRENT_HOST', socket.gethostname())])
        conf = json.load(
            open("/opt/ml/input/config/resourceconfig.json"))
        if group == None and len(conf["hosts"]) > 1:
            group = os.getenv('TRAINING_JOB_NAME')
        # Set secret variables
        if os.path.exists("secrets.env"):
            for line in open("secrets.env", "r"):
                key, val = line.strip().split('=', 1)
                os.environ[key] = val
    elif tf_config:
        cluster = tf_config.get('cluster')
        job_name = tf_config.get('task', {}).get('type')
        task_index = tf_config.get('task', {}).get('index')
        if job_name is not None and task_index is not None:
            # TODO: set run_id for resuming?
            run_id = cluster[job_name][task_index].rsplit(":")[0]
            if job_type == None:
                job_type = job_name
            if group == None and len(cluster.get("worker", [])) > 0:
                group = cluster[job_name][0].rsplit("-"+job_name, 1)[0]
    image = util.image_id_from_k8s()
    if image:
        os.environ[env.DOCKER] = image

    if not os.environ.get(env.SWEEP_ID):
        if project:
            os.environ[env.PROJECT] = project
        if entity:
            os.environ[env.ENTITY] = entity
    else:
        if entity and entity != os.environ.get(env.ENTITY):
            termwarn("Ignoring entity='{}' passed to wandb.init when running a sweep".format(entity))
        if project and project != os.environ.get(env.PROJECT):
            termwarn("Ignoring project='{}' passed to wandb.init when running a sweep".format(project))
    if save_code is not None:
        os.environ[env.SAVE_CODE]= str(save_code)
    if group:
        os.environ[env.RUN_GROUP] = group
    if job_type:
        os.environ[env.JOB_TYPE] = job_type
    if tags:
        if isinstance(tags, str):
            # People sometimes pass a string instead of an array of strings...
            tags = [tags]
        os.environ[env.TAGS] = ",".join(tags)
    if id:
        os.environ[env.RUN_ID] = id
        if name is None and resume != "must":
            # We do this because of https://github.com/wandb/core/issues/2170
            # to ensure that the run's name is explicitly set to match its
            # id. If we don't do this and the id is eight characters long, the
            # backend will set the name to a generated human-friendly value.
            #
            # In any case, if the user is explicitly setting `id` but not
            # `name`, their id is probably a meaningful string that we can
            # use to label the run.
            #
            # In the resume="must" case, we know we are resuming, so we should
            # make sure to not set the name because it would have been set with
            # the original run.
            #
            # TODO: handle "auto" resume by moving this logic later when we know
            # if there is a resume.
            name = os.environ.get(env.NAME, id)  # environment variable takes precedence over this.
    if name:
        os.environ[env.NAME] = name
    if notes:
        os.environ[env.NOTES] = notes
    if magic is not None and magic is not False:
        if isinstance(magic, dict):
            os.environ[env.MAGIC] = json.dumps(magic)
        elif isinstance(magic, str):
            os.environ[env.MAGIC] = magic
        elif isinstance(magic, bool):
            pass
        else:
            termwarn("wandb.init called with invalid magic parameter type", repeat=False)
        from wandb import magic_impl
        magic_impl.magic_install(init_args=init_args)
    if dir:
        os.environ[env.DIR] = dir
        util.mkdir_exists_ok(wandb_dir())
    if anonymous is not None:
        os.environ[env.ANONYMOUS] = anonymous
    if os.environ.get(env.ANONYMOUS, "never") not in ["allow", "must", "never"]:
        raise LaunchError("anonymous must be set to 'allow', 'must', or 'never'")

    resume_path = os.path.join(wandb_dir(), wandb_run.RESUME_FNAME)
    if resume == True:
        os.environ[env.RESUME] = "auto"
    elif resume in ("allow", "must", "never"):
        os.environ[env.RESUME] = resume
        if id:
            os.environ[env.RUN_ID] = id
    elif resume:
        os.environ[env.RESUME] = os.environ.get(env.RESUME, "allow")
        # TODO: remove allowing resume as a string in the future
        os.environ[env.RUN_ID] = id or resume
    elif os.path.exists(resume_path):
        os.remove(resume_path)
    if os.environ.get(env.RESUME) == 'auto' and os.path.exists(resume_path):
        if not os.environ.get(env.RUN_ID):
            os.environ[env.RUN_ID] = json.load(open(resume_path))["run_id"]

    # the following line is useful to ensure that no W&B logging happens in the user
    # process that might interfere with what they do
    # logging.basicConfig(format='user process %(asctime)s - %(name)s - %(levelname)s - %(message)s')

    # If a thread calls wandb.init() it will get the same Run object as
    # the parent. If a child process with distinct memory space calls
    # wandb.init(), it won't get an error, but it will get a result of
    # None.
    # This check ensures that a child process can safely call wandb.init()
    # after a parent has (only the parent will create the Run object).
    # This doesn't protect against the case where the parent doesn't call
    # wandb.init but two children do.
    if run or os.getenv(env.INITED):
        return run

    if __stage_dir__ is None:
        __stage_dir__ = "wandb"
        util.mkdir_exists_ok(wandb_dir())

    try:
        signal.signal(signal.SIGQUIT, _debugger)
    except (ValueError, AttributeError):
        pass

    try:
        run = wandb_run.Run.from_environment_or_defaults()
        _global_run_stack.append(run)
    except IOError as e:
        termerror('Failed to create run directory: {}'.format(e))
        raise LaunchError("Could not write to filesystem.")

    run.set_environment()

    def set_global_config(run):
        global config  # because we already have a local config
        config = run.config
    set_global_config(run)
    global summary
    summary = run.summary

    # set this immediately after setting the run and the config. if there is an
    # exception after this it'll probably break the user script anyway
    os.environ[env.INITED] = '1'

    if in_jupyter:
        _init_jupyter(run)
    elif run.mode == 'clirun':
        pass
    elif run.mode == 'run':
        api = InternalApi()
        # let init_jupyter handle this itself
        if not in_jupyter and not api.api_key:
            termlog(
                "W&B is a tool that helps track and visualize machine learning experiments")
            if force:
                termerror(
                    "No credentials found.  Run \"wandb login\" or \"wandb off\" to disable wandb")
            else:
                if util.prompt_api_key(api):
                    _init_headless(run)
                else:
                    termlog(
                        "No credentials found.  Run \"wandb login\" to visualize your metrics")
                    run.mode = "dryrun"
                    _init_headless(run, False)
        else:
            _init_headless(run)
    elif run.mode == 'dryrun':
        termlog(
            'Dry run mode, not syncing to the cloud.')
        _init_headless(run, False)
    else:
        termerror(
            'Invalid run mode "%s". Please unset WANDB_MODE.' % run.mode)
        raise LaunchError("The WANDB_MODE environment variable is invalid.")

    # set the run directory in the config so it actually gets persisted
    run.config.set_run_dir(run.dir)
    # we have re-read the config, add telemetry data
    telemetry_updated = run.config._telemetry_update()

    if sagemaker_config:
        run.config._update(sagemaker_config)
        allow_val_change = True
    if config or telemetry_updated:
        run.config._update(config,
                exclude_keys=config_exclude_keys,
                include_keys=config_include_keys,
                allow_val_change=allow_val_change,
                as_defaults=not allow_val_change)

    # Access history to ensure resumed is set when resuming
    run.history
    # Load the summary to support resuming
    run.summary.load()

    return run


tensorflow = util.LazyLoader('tensorflow', globals(), 'wandb.tensorflow')
tensorboard = util.LazyLoader('tensorboard', globals(), 'wandb.tensorboard')
jupyter = util.LazyLoader('jupyter', globals(), 'wandb.jupyter')
keras = util.LazyLoader('keras', globals(), 'wandb.keras')
fastai = util.LazyLoader('fastai', globals(), 'wandb.fastai')
docker = util.LazyLoader('docker', globals(), 'wandb.docker')
xgboost = util.LazyLoader('xgboost', globals(), 'wandb.xgboost')
lightgbm = util.LazyLoader('lightgbm', globals(), 'wandb.lightgbm')
catboost = util.LazyLoader('catboost', globals(), 'wandb.catboost')
gym = util.LazyLoader('gym', globals(), 'wandb.gym')
ray = util.LazyLoader('ray', globals(), 'wandb.ray')
sklearn = util.LazyLoader('sklearn', globals(), 'wandb.sklearn')


__all__ = ['init', 'config', 'summary', 'join', 'login', 'log', 'save', 'restore',
    'tensorflow', 'watch', 'types', 'tensorboard', 'jupyter', 'keras', 'fastai',
    'docker', 'lightgbm', 'catboost', 'xgboost', 'gym', 'ray', 'run', 'join', 'Image', 'Video',
    'Audio',  'Table', 'Html', 'Object3D', 'Molecule', 'Histogram', 'Graph', 'Api']



# -*- coding: utf-8 -*-
# Potential improvements:
#   - when add to pending_jobs, we should look and see if we already have a job
#     for this file, if so, don't bother adding it. We don't need more than
#     one pending
import collections
import os
import shutil
import tempfile as builtin_tempfile
import threading
import time
from six.moves import queue
import warnings
import tarfile
import multiprocessing

import wandb
import wandb.util
from wandb.compat import tempfile

from wandb.filesync import stats
from wandb.filesync import step_checksum
from wandb.filesync import step_prepare
from wandb.filesync import step_upload


def resolve_path(path):
    try:
        from pathlib import Path
        return str(Path(path).resolve())
    except:
        # Pathlib isn't present for python versions earlier than 3.3
        return os.path.realpath(path)


# Get rid of cleanup warnings in Python 2.7.
warnings.filterwarnings('ignore', 'Implicitly cleaning up', RuntimeWarning, 'wandb.compat.tempfile')


# Temporary directory for copies we make of some file types to
# reduce the probability that the file gets changed while we're
# uploading it.
TMP_DIR = tempfile.TemporaryDirectory('wandb')


class FilePusher(object):
    """Parallel file upload class.

    This manages uploading multiple files in parallel. It will restart a given file's
    upload job if it receives a notification that that file has been modified.
    The finish() method will block until all events have been processed and all
    uploads are complete.
    """

    MAX_UPLOAD_JOBS = 64

    def __init__(self, api):
        self._api = api

        self._tempdir = tempfile.TemporaryDirectory('wandb')

        self._stats = stats.Stats()

        self._incoming_queue = queue.Queue()
        self._event_queue = queue.Queue()

        self._step_checksum = step_checksum.StepChecksum(
            self._api, self._tempdir, self._incoming_queue, self._event_queue, self._stats)
        self._step_checksum.start()

        self._step_upload = step_upload.StepUpload(
            self._api, self._stats, self._event_queue, self.MAX_UPLOAD_JOBS)
        self._step_upload.start()

        # Holds refs to tempfiles if users need to make a temporary file that
        # stays around long enough for file pusher to sync
        # TODO(artifacts): maybe don't do this
        self._temp_file_refs = []


    def print_status(self, prefix=True):
        step = 0
        spinner_states = ['-', '\\', '|', '/']
        stop = False
        while True:
            if not self.is_alive():
                stop = True
            summary = self._stats.summary()
            line = ' %.2fMB of %.2fMB uploaded (%.2fMB deduped)\r' % (
                summary['uploaded_bytes'] / 1048576.0,
                summary['total_bytes'] / 1048576.0,
                summary['deduped_bytes'] / 1048576.0)
            line = spinner_states[step % 4] + line
            step += 1
            wandb.termlog(line, newline=False, prefix=prefix)
            if stop:
                break
            time.sleep(0.25)
        dedupe_fraction = summary['deduped_bytes'] / float(summary['total_bytes']) if summary['total_bytes'] > 0  else 0
        if dedupe_fraction > 0.01:
            wandb.termlog('✨ W&B sync reduced upload amount by %.1f%%             ' %
                          (dedupe_fraction * 100), prefix=prefix)
        # clear progress line.
        wandb.termlog(' ' * 79, prefix=prefix)

    def file_counts_by_category(self):
        return self._stats.file_counts_by_category()

    def file_changed(self, save_name, path, artifact_id=None, copy=True, use_prepare_flow=False, save_fn=None, digest=None):
        """Tell the file pusher that a file's changed and should be uploaded.

        Arguments:
            save_name: string logical location of the file relative to the run
                directory.
            path: actual string path of the file to upload on the filesystem.
        """
        # Tests in linux were failing because wandb-events.jsonl didn't exist
        if not os.path.exists(path) or not os.path.isfile(path):
            return
        if os.path.getsize(path) == 0:
            return

        event = step_checksum.RequestUpload(path, save_name, artifact_id, copy, use_prepare_flow, save_fn, digest)
        self._incoming_queue.put(event)

    def store_manifest_files(self, manifest, artifact_id, save_fn):
        event = step_checksum.RequestStoreManifestFiles(manifest, artifact_id, save_fn)
        self._incoming_queue.put(event)

    def named_temp_file(self, mode='w+b'):
        # get a named temp file that the file pusher with hold a reference to so it
        # doesn't get gc'd. Obviously, we shouldn't do this very much :). It's currently
        # used for artifact metadata.
        f = builtin_tempfile.NamedTemporaryFile(mode=mode, delete=False)
        self._temp_file_refs.append(f)
        return f

    def commit_artifact(self, artifact_id, on_commit=None):
        event = step_checksum.RequestCommitArtifact(artifact_id, on_commit)
        self._incoming_queue.put(event)

    def finish(self):
        self._incoming_queue.put(step_checksum.RequestFinish())

    def is_alive(self):
        return (self._step_checksum.is_alive()
            or self._step_upload.is_alive())
